# gencore_write_subr.py
from __future__ import absolute_import

from fortlab.resolver import statements, block_statements, typedecl_statements
from .verify_utils import kernel_verify_contains, kernel_verify_kgenutils, get_dtype_verifyname, check_class_derived

########################## Verbose level table ####################
#  Content #  level <1   #    level 1   #  level 2   # level >2   #
###################################################################
#  P/F     #    O        #      O       #     O      #    O       #
###################################################################
#  Stat    #    X        #      O       #     O      #    O       #
###################################################################
#  Ident   #             #              #            #            #
#  Result  #    X        #      X       #     O      #    O       #
###################################################################
#  Diff.   #             #              #            #            #
#  Result  #    X        #      O       #     O      #    O       #
###################################################################
#  Num.    #             #              #            #            #
#  Outtol. #             #              #            #            #
#  Diff.   #    X        #      O       #     O      #    O       #
#  Detail  #             #              #            #            #
###################################################################
#  Num.    #             #              #            #            #
#  Intol.  #             #              #            #            #
#  Diff.   #    X        #      X       #     O      #    O       #
#  Detail  #             #              #            #            #
###################################################################
#  Num.    #             #              #            #            #
#  Ident.  #    X        #      X       #     X      #    O       #
#  Detail  #             #              #            #            #
###################################################################
#  Comp.   #             #              #            #            #
#  Ident.  #    X        #      X       #     X      #    O       #
#  Result  #             #              #            #            #
###################################################################
#  Comp.   #             #              #            #            #
#  Diff.   #    X        #      X       #     O      #    O       #
#  Result  #             #              #            #            #
###################################################################
#  Comp.   #             #              #            #            #
#  Diff.   #    X        #      X       #     X      #    O       #
#  Detail  #             #              #            #            #
###################################################################

def create_verify_subr(subrname, entity_name, parent, var, stmt):

    def print_numarr_detail(parent, is_identical=False):
        attrs = {'items': ['count( var /= kgenref_var)', '" of "', 'size( var )', '" elements are different."']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"Average - kernel "', 'sum(var)/real(size(var))']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"Average - reference "', 'sum(kgenref_var)/real(size(kgenref_var))']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        if is_identical:
            attrs = {'items': ['"RMS of difference is "', '0']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['"Normalized RMS of difference is "', '0']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        else:
            attrs = {'items': ['"RMS of difference is "', 'rmsdiff']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['"Normalized RMS of difference is "', 'nrmsdiff']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

    def print_dtypearr_detail(parent, is_identical=False):

        attrs = {'items': [ '"    number of elements         : "', 'comp_check_status%numtotal']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': [ '"    identical                  : "', 'comp_check_status%numidentical']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': [ '"    not identical - out of tol.: "', 'comp_check_status%numouttol']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': [ '"    not identical - within tol.: "', 'comp_check_status%numintol']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

    def print_num_detail(parent, is_identical=False):
        if is_identical:
            attrs = {'items': ['"Difference is "', '0']}
        else:
            attrs = {'items': ['"Difference is "', 'diff']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

    def print_dummy_detail(parent, is_identical=False):
        attrs = {'items': ['"NOT IMPLEMENTED"']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

    def get_attrs(attrspec, allowed_attrs):
        attrspec = []
        for attr in stmt.attrspec:
            if any( attr.startswith(allowed_attr) for allowed_attr in allowed_attrs):
                attrspec.append(attr)
        return attrspec


    checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==subrname
    is_class_derived = check_class_derived(stmt)

    if  not part_has_node(parent, SUBP_PART, checks):

        checks = lambda n: n.kgen_isvalid and n.kgen_match_class==statements.Contains
        if not parent in kernel_verify_contains and not part_has_node(parent, CONTAINS_PART, checks):
            part_append_comment(parent, CONTAINS_PART, '')
            part_append_genknode(parent, CONTAINS_PART, statements.Contains)
            part_append_comment(parent, CONTAINS_PART, '')
            kernel_verify_contains.append(parent)

        checks = lambda n: n.kgen_isvalid and n.kgen_match_class==statements.Use and n.name=='kgen_utils_mod' and 'check_t' in n.items
        if not parent in kernel_verify_kgenutils and not part_has_node(parent, USE_PART, checks):
            attrs = {'name': 'kgen_utils_mod', 'isonly': True, 'items': ['check_t', 'kgen_init_check', 'kgen_init_verify',
                'kgen_tolerance', 'kgen_minvalue', 'kgen_verboselevel', 'CHECK_IDENTICAL', 'CHECK_IN_TOL', 'CHECK_OUT_TOL']}
            part_append_genknode(parent, USE_PART, statements.Use, attrs=attrs)
            kernel_verify_kgenutils.append(parent)

        part_append_comment(parent, SUBP_PART, 'verify state subroutine for %s'%subrname)
        attrs = {'prefix': 'RECURSIVE', 'name': subrname, 'args': ['varname', 'check_status', 'var', 'kgenref_var']}
        subrobj = part_append_genknode(parent, SUBP_PART, block_statements.Subroutine, attrs=attrs)
        part_append_comment(parent, SUBP_PART, '')

        # varname
        attrs = {'type_spec': 'CHARACTER', 'attrspec': ['INTENT(IN)'], 'selector':('*', None), 'entity_decls': ['varname']}
        part_append_genknode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

        # check_status
        attrs = {'type_spec': 'TYPE', 'attrspec': ['INTENT(INOUT)'], 'selector':(None, 'check_t'), 'entity_decls': ['check_status']}
        part_append_genknode(subrobj, DECL_PART, typedecl_statements.Type, attrs=attrs)

        # variable
        type_spec = stmt.__class__.__name__.upper()
        selector = stmt.selector
        var_class = stmt.__class__
        attrspec = get_attrs(stmt.attrspec, ['pointer', 'allocatable']) + ['INTENT(IN)']
        if var.is_array():
            attrspec.append('DIMENSION(%s)'%','.join(':'*var.rank))
            if stmt.is_derived() or is_class_derived:
                # comp_check_status
                attrs = {'type_spec': 'TYPE', 'selector':(None, 'check_t'), 'entity_decls': ['comp_check_status']}
                part_append_genknode(subrobj, DECL_PART, typedecl_statements.Type, attrs=attrs)
            if stmt.is_derived():
                type_spec = 'TYPE'
                selector = (None, stmt.name)
                var_class = typedecl_statements.Type
        else:
            if (stmt.is_derived() or is_class_derived) and not var.is_pointer():
                raise Exception('Non array or not pointer of derived type should not be processed here.')
        attrs = {'type_spec': type_spec, 'attrspec': attrspec, 'selector':selector, 'entity_decls': ['var', 'kgenref_var']}
        part_append_genknode(subrobj, DECL_PART, var_class, attrs=attrs)

        # check result
        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['check_result']}
        part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        # print check result
        attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['is_print = .FALSE.']}
        part_append_genknode(subrobj, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        part_append_comment(subrobj, DECL_PART, '')

        part_append_comment(subrobj, EXEC_PART, '')

        ifallocobj = None
        ifassocobj = None
        if var.is_allocatable():
            attrs = {'expr': 'ALLOCATED(var)'}
            ifallocobj = part_append_genknode(subrobj, EXEC_PART, block_statements.IfThen, attrs=attrs)
            if var.is_pointer():
                attrs = {'expr': 'ASSOCIATED(var)'}
                ifassocobj = part_append_genknode(ifallocobj, EXEC_PART, block_statements.IfThen, attrs=attrs)
        elif var.is_pointer():
            attrs = {'expr': 'ASSOCIATED(var)'}
            ifassocobj = part_append_genknode(subrobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        if ifassocobj: topobj = ifassocobj
        elif ifallocobj: topobj = ifallocobj
        else: topobj = subrobj

        attrs = {'variable': 'check_status%numTotal', 'sign': '=', 'expr': 'check_status%numTotal + 1'}
        part_append_genknode(topobj, EXEC_PART, statements.Assignment, attrs=attrs)
        part_append_comment(topobj, EXEC_PART, '')

        if var.is_array():

            # index
            attrs = {'type_spec': 'INTEGER', 'entity_decls': [ 'idx%d'%(d+1) for d in range(var.rank)]}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            dim_shape = ','.join(':'*var.rank)
            get_size = ','.join(['SIZE(var,dim=%d)'%(dim+1) for dim in range(var.rank)])

            if stmt.is_derived() or is_class_derived:

                callname = None
                for uname, req in stmt.unknowns.items():
                    #if uname.firstpartname()==stmt.name:
                    if ( is_class_derived and uname.firstpartname()==stmt.selector[1]) or uname.firstpartname()==stmt.name:
                        if len(req.res_stmts)>0:
                            callname = get_dtype_verifyname(req.res_stmts[0])
                            break

                attrs = {'designator': 'kgen_init_check', 'items': ['comp_check_status', 'rank=check_status%rank']}
                part_append_genknode(topobj, EXEC_PART, statements.Call, attrs=attrs)

                if callname:
                    pobj = topobj
                    doobjs = []
                    for d in range(var.rank):
                        attrs = {'loopcontrol': ' idx%(d)d=LBOUND(var,%(d)d), UBOUND(var,%(d)d)'%{'d':d+1}}
                        doobj = part_append_genknode(pobj, EXEC_PART, block_statements.Do, attrs=attrs)
                        doobjs.append(doobj)
                        pobj = doobj 

                    # call verify subr
                    indexes = ','.join([ 'idx%d'%(r+1) for r in range(var.rank)])
                    attrs = {'designator': callname, 'items': ['trim(adjustl(varname))', 'comp_check_status', \
                        'var(%s)'%indexes, 'kgenref_var(%s)'%indexes]}
                    part_append_genknode(doobjs[-1], EXEC_PART, statements.Call, attrs=attrs)

                    attrs = {'expr': 'comp_check_status%numTotal == comp_check_status%numIdentical'}
                    ifidobj = part_append_genknode(topobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'variable': 'check_status%numIdentical', 'sign': '=', 'expr': 'check_status%numIdentical + 1'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'kgen_verboseLevel > 1'}
                    ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'expr': 'check_status%rank == 0'}
                    ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'items': ['trim(adjustl(varname))','" is IDENTICAL."']}
                    part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IDENTICAL'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'comp_check_status%numOutTol > 0'}
                    part_append_genknode(ifidobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

                    attrs = {'variable': 'check_status%numOutTol', 'sign': '=', 'expr': 'check_status%numOutTol + 1'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'kgen_verboseLevel > 0'}
                    ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'expr': 'check_status%rank == 0'}
                    ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'items': ['trim(adjustl(varname))','" is NOT IDENTICAL(out of tolerance)."']}
                    part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_OUT_TOL'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'comp_check_status%numInTol > 0'}
                    part_append_genknode(ifidobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

                    attrs = {'variable': 'check_status%numInTol', 'sign': '=', 'expr': 'check_status%numInTol + 1'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'kgen_verboseLevel > 0'}
                    ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'expr': 'check_status%rank == 0'}
                    ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'items': ['trim(adjustl(varname))','" is NOT IDENTICAL(within tolerance)."']}
                    part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IN_TOL'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)
            else: # intrinsic type

                attrs = {'type_spec': 'INTEGER', 'entity_decls': ['n']}
                part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

                if stmt.name=='logical':
                    attrs = {'expr': 'ALL(var .EQV. kgenref_var)'}
                else:
                    attrs = {'expr': 'ALL(var == kgenref_var)'}

                ifidobj = part_append_genknode(topobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                attrs = {'variable': 'check_status%numIdentical', 'sign': '=', 'expr': 'check_status%numIdentical + 1'}
                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                attrs = {'expr': 'kgen_verboseLevel > 1'}
                ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                attrs = {'expr': 'check_status%rank == 0'}
                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                attrs = {'items': ['trim(adjustl(varname))','" is IDENTICAL."']}
                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IDENTICAL'}
                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                part_append_genknode(ifidobj, EXEC_PART, statements.Else)

                if stmt.is_numeric():

                    # typececls
                    attrs = {'type_spec': stmt.name, 'selector':stmt.selector, 'entity_decls': ['nrmsdiff', 'rmsdiff']}
                    part_append_genknode(subrobj, DECL_PART, stmt.__class__, attrs=attrs)

                    attrs = {'type_spec': stmt.name, 'attrspec': ['ALLOCATABLE'], 'selector':stmt.selector, 'entity_decls': \
                        ['buf1(%s)'%dim_shape,'buf2(%s)'%dim_shape]}
                    part_append_genknode(subrobj, DECL_PART, stmt.__class__, attrs=attrs)

                    attrs = {'items': ['buf1(%s)'%get_size]}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Allocate, attrs=attrs)

                    attrs = {'items': ['buf2(%s)'%get_size]}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Allocate, attrs=attrs)

                    #attrs = {'variable': 'n', 'sign': '=', 'expr': 'COUNT(var /= kgenref_var)'}
                    attrs = {'variable': 'n', 'sign': '=', 'expr': 'SIZE(var)'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'ABS(kgenref_var) > kgen_minvalue'}
                    whereobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.Where, attrs=attrs)

                    attrs = {'variable': 'buf1', 'sign': '=', 'expr': '((var-kgenref_var)/kgenref_var)**2'}
                    part_append_genknode(whereobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'variable': 'buf2', 'sign': '=', 'expr': '(var-kgenref_var)**2'}
                    part_append_genknode(whereobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    part_append_genknode(whereobj, EXEC_PART, statements.ElseWhere)

                    attrs = {'variable': 'buf1', 'sign': '=', 'expr': '(var-kgenref_var)**2'}
                    part_append_genknode(whereobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'variable': 'buf2', 'sign': '=', 'expr': 'buf1'}
                    part_append_genknode(whereobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'variable': 'nrmsdiff', 'sign': '=', 'expr': 'SQRT(SUM(buf1)/DBLE(n))'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'variable': 'rmsdiff', 'sign': '=', 'expr': 'SQRT(SUM(buf2)/DBLE(n))'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    #attrs = {'expr': 'nrmsdiff > kgen_tolerance'}
                    attrs = {'expr': 'rmsdiff > kgen_tolerance'}
                    ifvobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'variable': 'check_status%numOutTol', 'sign': '=', 'expr': 'check_status%numOutTol + 1'}
                    part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'kgen_verboseLevel > 0'}
                    ifvlobj = part_append_genknode(ifvobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'expr': 'check_status%rank == 0'}
                    ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'items': ['trim(adjustl(varname))','" is NOT IDENTICAL(out of tolerance)."']}
                    part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_OUT_TOL'}
                    part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    part_append_genknode(ifvobj, EXEC_PART, statements.Else)

                    attrs = {'variable': 'check_status%numInTol', 'sign': '=', 'expr': 'check_status%numInTol + 1'}
                    part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'kgen_verboseLevel > 1'}
                    ifvlobj = part_append_genknode(ifvobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'expr': 'check_status%rank == 0'}
                    ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'items': ['trim(adjustl(varname))','" is NOT IDENTICAL(within tolerance)."']}
                    part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IN_TOL'}
                    part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)
                else: # not numerical

                    if stmt.name=='logical':
                        attrs = {'variable': 'n', 'sign': '=', 'expr': 'COUNT(var .NEQV. kgenref_var)'}
                    else:
                        attrs = {'variable': 'n', 'sign': '=', 'expr': 'COUNT(var /= kgenref_var)'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'variable': 'check_status%numOutTol', 'sign': '=', 'expr': 'check_status%numOutTol + 1'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'kgen_verboseLevel > 0'}
                    ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'expr': 'check_status%rank == 0'}
                    ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'items': ['trim(adjustl(varname))','" is NOT IDENTICAL(out of tolerance)."']}
                    part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_OUT_TOL'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)
        else: # scalar
            if not stmt.is_derived() or is_class_derived:
                # diff
                attrs = {'type_spec': stmt.name, 'selector':stmt.selector, 'entity_decls': ['diff']}
                part_append_genknode(subrobj, DECL_PART, stmt.__class__, attrs=attrs)

                if stmt.name=='logical':
                    attrs = {'expr': 'var .EQV. kgenref_var'}
                else:
                    attrs = {'expr': 'var == kgenref_var'}
                ifidobj = part_append_genknode(topobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                attrs = {'variable': 'check_status%numIdentical', 'sign': '=', 'expr': 'check_status%numIdentical + 1'}
                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                attrs = {'expr': 'kgen_verboseLevel > 1'}
                ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                attrs = {'expr': 'check_status%rank == 0'}
                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                attrs = {'items': ['trim(adjustl(varname))','" is IDENTICAL."']}
                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IDENTICAL'}
                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                part_append_genknode(ifidobj, EXEC_PART, statements.Else)

                if stmt.is_numeric():

                    attrs = {'variable': 'diff', 'sign': '=', 'expr': 'ABS(var - kgenref_var)'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'diff <= kgen_tolerance'}
                    ifvobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'variable': 'check_status%numInTol', 'sign': '=', 'expr': 'check_status%numInTol + 1'}
                    part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'kgen_verboseLevel > 1'}
                    ifvlobj = part_append_genknode(ifvobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'expr': 'check_status%rank == 0'}
                    ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'items': ['trim(adjustl(varname))','" is NOT IDENTICAL(within tolerance)."']}
                    part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IN_TOL'}
                    part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    part_append_genknode(ifvobj, EXEC_PART, statements.Else)

                    attrs = {'variable': 'check_status%numOutTol', 'sign': '=', 'expr': 'check_status%numOutTol + 1'}
                    part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'kgen_verboseLevel > 0'}
                    ifvlobj = part_append_genknode(ifvobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'expr': 'check_status%rank == 0'}
                    ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'items': ['trim(adjustl(varname))','" is NOT IDENTICAL(out of tolerance)."']}
                    part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_OUT_TOL'}
                    part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)
                else: # not numeric

                    attrs = {'variable': 'check_status%numOutTol', 'sign': '=', 'expr': 'check_status%numOutTol + 1'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'expr': 'kgen_verboseLevel > 1'}
                    ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'expr': 'check_status%rank == 0'}
                    ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'items': ['trim(adjustl(varname))','" is NOT IDENTICAL."']}
                    part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_OUT_TOL'}
                    part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

        print_detail = print_dummy_detail
        if var.is_array(): # array
            if stmt.is_derived() or is_class_derived:
                print_detail = print_dtypearr_detail
            else:
                if stmt.is_numeric():
                    print_detail = print_numarr_detail
                else:
                    pass
        else:
            if stmt.is_derived() or is_class_derived:
                pass
            else:
                if stmt.is_numeric():
                    print_detail = print_num_detail
                else:
                    pass

        attrs = {'expr': 'check_result == CHECK_IDENTICAL'}
        ifchkobj = part_append_genknode(topobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'expr': 'kgen_verboseLevel > 2'}
        iflevel3obj = part_append_genknode(ifchkobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'expr': 'check_status%rank == 0'}
        ifrank = part_append_genknode(iflevel3obj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        print_detail(ifrank, is_identical=True)

        attrs = {'expr': 'check_result == CHECK_OUT_TOL'}
        part_append_genknode(ifchkobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

        attrs = {'expr': 'kgen_verboseLevel > 0'}
        iflevel0obj = part_append_genknode(ifchkobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'expr': 'check_status%rank == 0'}
        ifrank = part_append_genknode(iflevel0obj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        print_detail(ifrank)

        attrs = {'expr': 'check_result == CHECK_IN_TOL'}
        part_append_genknode(ifchkobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

        attrs = {'expr': 'kgen_verboseLevel > 1'}
        iflevel1obj = part_append_genknode(ifchkobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'expr': 'check_status%rank == 0'}
        ifrank = part_append_genknode(iflevel1obj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        print_detail(ifrank)

    part_append_comment(topobj, EXEC_PART, '')

    # create public stmt
#    if parent.kgen_match_class in [ block_statements.Program, block_statements.Module ]:
#        attrs = {'items': [subrname]}
#        part_append_genknode(parent, DECL_PART, statements.Public, attrs=attrs)



"""
-----
Permission to use, modify, and distribute this software is given under the
terms of the NumPy License. See http://scipy.org.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Author: Pearu Peterson <pearu@cens.ioc.ee>
Created: May 2006
-----
"""

__all__ = ['Statement','BeginStatement','EndStatement', 'Variable',
           'AttributeHolder','ProgramBlock']

import re
import sys
import copy
from collections import OrderedDict
from fortlab.resolver.readfortran import Line, Comment
#from numpy.distutils.misc_util import yellow_text, red_text # KGEN deletion
from fortlab.resolver.utils import split_comma, specs_split_comma, is_int_literal_constant
from fortlab.resolver.utils import classes

#logger = logging.getlogger('fparser') # KGEN deletion

# start of KGEN addition
#import logging
#logger = logging.getLogger('kgen')

from fortlab.kgutils import KGName, ProgramException, traverse, logger
from fortlab.resolver.kgintrinsics import Intrinsic_Procedures

class KGen_Resolver(object):
    def __init__(self, name):
        self.name = name

    def tokgen():
        return ''

    # save names that this self resolved
    def add_geninfo(self, uname, request):

        if uname is None or request is None: return

        if not hasattr(self, 'geninfo'):
            self.geninfo = OrderedDict()
        if request.gentype not in self.geninfo:
            self.geninfo[request.gentype] = []
        if (uname, request) not in self.geninfo[request.gentype]:
            self.geninfo[request.gentype].append((uname, request))

    def get_variable(self, name):
        return None

class ImplicitRule_Resolver(KGen_Resolver):
    pass

class IntrinsicProcedure_Resolver(KGen_Resolver):
    pass

def is_except(name, stmt, excepts):
    if not name or not stmt: return False

    namelist = [a.name for a in stmt.ancestors()]
    namelist.append(name)
    #exceptlist = Config.search['except']

    #for elist in exceptlist:
    for elist in excepts:
        elist_split = elist.split(':')
        same = True
        for i in range(min(len(namelist), len(elist_split))):
            if namelist[len(namelist)-i-1]!=elist_split[len(elist_split)-i-1]:
                same = False
                break
        if same: return True

    return False

# end of KGEN addition

class AttributeHolder(object):
    # copied from symbolic.base module
    """
    Defines a object with predefined attributes. Only those attributes
    are allowed that are specified as keyword arguments of a constructor.
    When an argument is callable then the corresponding attribute will
    be read-only and set by the value the callable object returns.
    """
    def __init__(self, **kws):
        self._attributes = {}
        self._readonly = []
        for k,v in kws.items():
            self._attributes[k] = v
            if callable(v):
                self._readonly.append(k)
        return

    def __getattr__(self, name):
        if name not in self._attributes:
            raise AttributeError('%s instance has no attribute %r, '\
                  'expected attributes: %s' \
                  % (self.__class__.__name__,name,
                     ','.join(self._attributes.keys())))
        value = self._attributes[name]
        if callable(value):
            value = value()
            self._attributes[name] = value
        return value

    def __setattr__(self, name, value):
        if name in ['_attributes','_readonly']:
            self.__dict__[name] = value
            return
        if name in self._readonly:
            raise AttributeError('%s instance attribute %r is readonly' \
                  % (self.__class__.__name__, name))
        if name not in self._attributes:
            raise AttributeError('%s instance has no attribute %r, '\
                  'expected attributes: %s' \
                  % (self.__class__.__name__,name,','.join(self._attributes.keys())))
        self._attributes[name] = value

    def isempty(self):
        for k in self._attributes.keys():
            v = getattr(self,k)
            if v: return False
        return True

    def __repr__(self): return self.torepr()

    def torepr(self, depth=-1, tab = ''):
        if depth==0: return tab + self.__class__.__name__
        l = [self.__class__.__name__+':']
        ttab = tab + '    '
        for k in self._attributes.keys():
            v = getattr(self,k)
            if v:
                if isinstance(v,list):
                    l.append(ttab + '%s=<%s-list>' % (k,len(v)))
                elif isinstance(v,dict):
                    l.append(ttab + '%s=<dict with keys %s>' % (k,v.keys()))
                else:
                    l.append(ttab + '%s=<%s>' % (k,type(v)))
        return '\n'.join(l)

    def todict(self):
        d = {}
        for k in self._attributes.keys():
            v = getattr(self, k)
            d[k] = v
        return d

def get_base_classes(cls):
    bases = ()
    for c in cls.__bases__:
        bases += get_base_classes(c)
    return bases + cls.__bases__ + (cls,)

class Variable(object):
#class Variable(metaclass=classes):
    """
    Variable instance has attributes:
      name
      typedecl
      dimension
      attributes
      intent
      parent - Statement instances defining the variable
    """

    #__metaclass__ = classes
    
    def __init__(self, parent, name):
        self.parent = parent
        self.parents = [parent]
        self.name = name
        self.typedecl = None
        self.dimension = None
        self.bounds = None
        self.length = None
        self.attributes = []
        self.intent = None
        self.bind = []
        self.check = []
        self.init = None

        # after calling analyze the following additional attributes are set:
        # .is_array:
        #    rank
        #    shape
        return

    def __repr__(self):
        l = []
        for a in ['name','typedecl','dimension','bounds','length','attributes','intent','bind','check','init']:
            v = getattr(self,a)
            if v:
                l.append('%s=%r' % (a,v))
        return 'Variable: ' + ', '.join(l)

    def get_bit_size(self):
        typesize = self.typedecl.get_bit_size()
        if self.is_pointer():
            # The size of pointer descriptor is compiler version dependent. Read:
            #   http://www.nersc.gov/vendor_docs/intel/f_ug1/pgwarray.htm
            #   https://www.cca-forum.org/pipermail/cca-fortran/2003-February/000123.html
            #   https://www.cca-forum.org/pipermail/cca-fortran/2003-February/000122.html
            # On sgi descriptor size may be 128+ bits!
            if self.is_array():
                wordsize = 4 # XXX: on a 64-bit system it is 8.
                rank = len(self.bounds or self.dimension)
                return 6 * wordsize + 12 * rank
            return typesize
        if self.is_array():
            size = reduce(lambda x,y:x*y,self.bounds or self.dimension,1)
            if self.length:
                size *= self.length
            return size * typesize
        if self.length:
            return self.length * typesize
        return typesize

    def get_typedecl(self):
        if self.typedecl is None:
            self.set_type(self.parent.get_type(self.name))
        return self.typedecl

    def add_parent(self, parent):
        if id(parent) not in map(id, self.parents):
            self.parents.append(parent)
        self.parent = parent
        return

    def set_type(self, typedecl):
        if self.typedecl is not None:
            if not self.typedecl==typedecl:
                self.parent.warning(\
                    'variable %r already has type %s,'\
                    ' resetting to %s' \
                    % (self.name, self.typedecl.tostr(),typedecl.tostr()))
        assert typedecl is not None
        self.typedecl = typedecl
        return

    def set_init(self, expr):
        if self.init is not None:
            if not self.init==expr:
                self.parent.warning(\
                    'variable %r already has initialization %r, '\
                    ' resetting to %r' % (self.name, self.expr, expr))
        self.init = expr
        return

    def set_dimension(self, dims):
        dims = [tuple(dim.split(':')) for dim in dims]
        dims = [tuple(map(str.strip, dim)) for dim in dims]
        if self.dimension is not None:
            if not self.dimension==dims:
                self.parent.warning(\
                    'variable %r already has dimension %r, '\
                    ' resetting to %r' % (self.name, self.dimension, dims))
        self.dimension = dims
        return

    def set_bounds(self, bounds):
        if self.bounds is not None:
            if not self.bounds==bounds:
                self.parent.warning(\
                    'variable %r already has bounds %r, '\
                    ' resetting to %r' % (self.name, self.bounds, bounds))
        self.bounds = bounds
        return

    def set_length(self, length):
        if self.length is not None:
            if not self.length==length:
                self.parent.warning(\
                    'variable %r already has length %r, '\
                    ' resetting to %r' % (self.name, self.length, length))
        self.length = length
        return

    known_intent_specs = ['IN','OUT','INOUT','CACHE','HIDE', 'COPY',
                          'OVERWRITE', 'CALLBACK', 'AUX', 'C', 'INPLACE',
                          'OUT=']

    def set_intent(self, intent):
        if self.intent is None:
            self.intent = []
        for i in intent:
            if i not in self.intent:
                if i not in self.known_intent_specs:
                    self.parent.warning('unknown intent-spec %r for %r'\
                                        % (i, self.name))
                self.intent.append(i)
        return

    known_attributes = ['PUBLIC', 'PRIVATE', 'ALLOCATABLE', 'ASYNCHRONOUS',
                        'EXTERNAL', 'INTRINSIC', 'OPTIONAL', 'PARAMETER',
                        'POINTER', 'PROTECTED', 'SAVE', 'TARGET', 'VALUE',
                        'VOLATILE', 'REQUIRED', 'DEFERRED', 'CONTINUOUS'] # KGEN addition
                        #'VOLATILE', 'REQUIRED'] # KGEN deletion

    def is_intent_in(self):
        #if not self.intent: return True # KGEN deletion
        if not self.intent: return False # KGEN addition
        # start of KGEN    
        #if not self.intent:
        #    anc = self.parent.ancestors()[-1]
        #    if hasattr(anc, 'args') and self.name in anc.args: return True
        #    else: return False
        # end of KGEN
        if 'HIDE' in self.intent: return False
        if 'INPLACE' in self.intent: return False
        if 'IN' in self.intent: return True
        if 'OUT' in self.intent: return False
        if 'INOUT' in self.intent: return False
        if 'OUTIN' in self.intent: return False
        return True

    def is_intent_inout(self):
        #if not self.intent: return False # KGEN deletion
        # start of KGEN    
        if not self.intent:
            anc = self.parent.ancestors()[-1]
            if hasattr(anc, 'args') and self.name in anc.args: return True
            else: return False
        # end of KGEN

        if 'INOUT' in self.intent:
            if 'IN' in self.intent or 'HIDE' in self.intent or 'INPLACE' in self.intent:
                self.warning('INOUT ignored in INPUT(%s)' % (', '.join(self.intent)))
                return False
            return True
        return False

    def is_intent_hide(self):
        if not self.intent: return False
        if 'HIDE' in self.intent: return True
        if 'OUT' in self.intent:
            return 'IN' not in self.intent and 'INPLACE' not in self.intent and 'INOUT' not in self.intent
        return False

    def is_intent_inplace(self): return self.intent and 'INPLACE' in self.intent
    def is_intent_out(self): return  self.intent and 'OUT' in self.intent
    def is_intent_c(self): return  self.intent and 'C' in self.intent
    def is_intent_cache(self): return  self.intent and 'CACHE' in self.intent
    def is_intent_copy(self): return  self.intent and 'COPY' in self.intent
    def is_intent_overwrite(self): return  self.intent and 'OVERWRITE' in self.intent
    def is_intent_callback(self): return  self.intent and 'CALLBACK' in self.intent
    def is_intent_aux(self): return  self.intent and 'AUX' in self.intent

    def is_private(self):
        from fortlab.resolver.block_statements import BeginSource
        if 'PUBLIC' in self.attributes: return False
        if 'PRIVATE' in self.attributes: return True
        # start of KGEN
#        if isinstance(self.parent.parent, BeginSource):
#            return self.parent.check_private(self.name)
#        else:
#            return self.parent.parent.check_private(self.name)
        if hasattr(self.parent.parent, 'check_private'):
            return self.parent.parent.check_private(self.name)
        else:
            return False
        # end of KGEN
        # return self.parent.parent.check_private(self.name) # KGEN deletion
    def is_public(self): return not self.is_private()

    def is_allocatable(self): return 'ALLOCATABLE' in self.attributes
    def is_external(self): return 'EXTERNAL' in self.attributes
    def is_intrinsic(self): return 'INTRINSIC' in self.attributes
    def is_parameter(self): return 'PARAMETER' in self.attributes
    def is_optional(self): return 'OPTIONAL' in self.attributes and 'REQUIRED' not in self.attributes and not self.is_intent_hide()
    def is_required(self): return self.is_optional() and not self.is_intent_hide()
    def is_pointer(self): return 'POINTER' in self.attributes
    def is_contiguous(self): return 'CONTIGUOUS' in self.attributes # KGen addition

    def is_array(self): return not not (self.bounds or self.dimension)
    def is_scalar(self): return not self.is_array()

    def update(self, *attrs):
        attributes = self.attributes
        if len(attrs)==1 and isinstance(attrs[0],(tuple,list)):
            attrs = attrs[0]
        for attr in attrs:
            lattr = attr.lower()
            uattr = attr.upper()
            if lattr.startswith('dimension'):
                assert self.dimension is None, str(self.dimension) + ", " + str(attr)
                l = attr[9:].lstrip()
                assert l[0]+l[-1]=='()',str(l)
                self.set_dimension(split_comma(l[1:-1].strip(), self.parent.item))
                continue
            if lattr.startswith('intent'):
                l = attr[6:].lstrip()
                assert l[0]+l[-1]=='()',str(l)
                self.set_intent(specs_split_comma(l[1:-1].strip(),
                                                  self.parent.item, upper=True))
                continue
            if lattr.startswith('bind'):
                l = attr[4:].lstrip()
                assert l[0]+l[-1]=='()',str(l)
                self.bind = specs_split_comma(l[1:-1].strip(), self.parent.item,
                                              upper = True)
                continue
            if lattr.startswith('check'):
                l = attr[5:].lstrip()
                assert l[0]+l[-1]=='()',str(l)
                self.check.extend(split_comma(l[1:-1].strip(), self.parent.item))
                continue
            if uattr not in attributes:
                if uattr not in self.known_attributes:
                    self.parent.warning('unknown attribute %r' % (attr))
                attributes.append(uattr)
        return

    def __str__(self):
        s = ''
        typedecl = self.get_typedecl()
        if typedecl is not None:
            s += typedecl.tostr() + ' '
        a = self.attributes[:]
        if self.dimension is not None:
            a.append('DIMENSION(%s)' % (', '.join([':'.join(spec) for spec in self.dimension])))
        if self.intent is not None:
            a.append('INTENT(%s)' % (', '.join(self.intent)))
        if self.bind:
            a.append('BIND(%s)' % (', '.join(self.bind)))
        if self.check:
            a.append('CHECK(%s)' % (', '.join(self.check)))
        if a:
            s += ', ' + ', '.join(a) + ' :: '
        s += self.name
        if self.bounds:
            s += '(%s)' % (', '.join([':'.join(spec) for spec in self.bounds]))
        if self.length:
            if is_int_literal_constant(self.length):
                s += '*%s' % (self.length)
            else:
                s += '*(%s)' % (self.length)
        if self.init:
            s += ' = ' + self.init
        return s

    def get_array_spec(self):
        assert self.is_array(),'array_spec is available only for arrays'
        if self.bounds:
            if self.dimension:
                self.parent.warning('both bounds=%r and dimension=%r are defined, ignoring dimension.' % (self.bounds, self.dimension))
            array_spec = self.bounds
        else:
            array_spec = self.dimension
        return array_spec

    def is_deferred_shape_array(self):
        if not self.is_array(): return False
        return self.is_allocatable() or self.is_pointer()

    def is_assumed_size_array(self):
        if not self.is_array(): return False
        return self.get_array_spec()[-1][-1]=='*'

    def is_assumed_shape_array(self):
        if not self.is_array(): return False
        if self.is_deferred_shape_array(): return False
        for spec in self.get_array_spec():
            if not spec[-1]: return True
        return False

    def is_explicit_shape_array(self):
        if not self.is_array(): return False
        if self.is_deferred_shape_array(): return False
        for spec in self.get_array_spec():
            if not spec[-1] or spec[-1] == '*': return False
        return True

    def is_allocatable_array(self):
        return self.is_array() and self.is_allocatable()

    def is_array_pointer(self):
        return self.is_array() and self.is_pointer()

    def analyze(self):
        typedecl = self.get_typedecl()
        if self.is_array():
            array_spec = self.get_array_spec()
            self.rank = len(array_spec)
            if self.is_deferred_shape_array(): # a(:,:)
                pass
            elif self.is_explicit_shape_array():
                shape = []
                for spec in array_spec:
                    if len(spec)==1:
                        shape.append(spec[0])
                    else:
                        try:
                            n = int(spec[1]) - int(spec[0])
                        except ValueError:
                            n = '(%s)-(%s)' % (spec[1], spec[0])
                        shape.append(str(n))
                self.shape = shape
        return

    def error(self, message):
        return self.parent.error(message)
    def warning(self, message):
        return self.parent.warning(message)
    def info(self, message):
        return self.parent.info(message)

class ProgramBlock(object):
#class ProgramBlock(metaclass=classes):

    pass
    #__metaclass__ = classes

class Statement(object):
#class Statement(metaclass=classes):
    """
    Statement instance has attributes:
      parent  - Parent BeginStatement or FortranParser instance
      item    - Line instance containing the statement line
      isvalid - boolean, when False, the Statement instance will be ignored
    """
    #__metaclass__ = classes

    modes = ['free','fix','f77','pyf']
    _repr_attr_names = []

    def __init__(self, parent, item):
        self.parent = parent
        if item is not None:
            self.reader = item.reader
        else:
            self.reader = parent.reader
        self.top = getattr(parent,'top',None) # the top of statement tree
        self.item = item

        if isinstance(parent, ProgramBlock):
            self.programblock = parent
        elif isinstance(self, ProgramBlock):
            self.programblock = self
        elif hasattr(parent,'programblock'):
            self.programblock = parent.programblock
        else:
            #self.warning('%s.programblock attribute not set.' % (self.__class__.__name__))
            pass

        # when a statement instance is constructed by error, set isvalid to False
        self.isvalid = True
        # when a statement should be ignored, set ignore to True
        self.ignore = False

        # attribute a will hold analyze information.
        a_dict = {}
        for cls in get_base_classes(self.__class__):
            if hasattr(cls,'a'):
                a_dict.update(copy.deepcopy(cls.a.todict()))
        self.a = AttributeHolder(**a_dict)
        if hasattr(self.__class__,'a'):
            assert self.a is not self.__class__.a

        self.process_item()

        return

    def __repr__(self):
        return self.torepr()

    def torepr(self, depth=-1,incrtab=''):
        tab = incrtab + self.get_indent_tab()
        clsname = self.__class__.__name__
        #l = [tab + yellow_text(clsname)] # KGEN deletion
        l = [tab + clsname] # KGEN addition
        if depth==0:
            return '\n'.join(l)
        ttab = tab + '  '
        for n in self._repr_attr_names:
            attr = getattr(self, n, None)
            if not attr: continue
            if hasattr(attr, 'torepr'):
                r = attr.torepr(depth-1,incrtab)
            else:
                r = repr(attr)
            l.append(ttab + '%s=%s' % (n, r))
        if self.item is not None: l.append(ttab + 'item=%r' % (self.item))
        if not self.isvalid: l.append(ttab + 'isvalid=%r' % (self.isvalid))
        if self.ignore: l.append(ttab + 'ignore=%r' % (self.ignore))
        if not self.a.isempty():
            l.append(ttab + 'a=' + self.a.torepr(depth-1,incrtab+'  ').lstrip())
        return '\n'.join(l)

    def get_indent_tab(self,deindent=False,isfix=None):
        if isfix is None: isfix = self.reader.isfixed
        if isfix:
            tab = ' '*6
        else:
            tab = ''
        p = self.parent
        while isinstance(p, Statement):
            tab += '  '
            p = p.parent
        if deindent:
            tab = tab[:-2]
        label = getattr(self.item, 'label', None)
        if label is None:
            return tab
        s = str(label)
        if isfix:
            s = ' '+s
        tab = tab[len(s):]
        if not tab: tab = ' '
        tab = s + tab
        return tab

    def __str__(self):
        return self.tofortran()

    def asfix(self):
        lines = []
        for line in self.tofortran(isfix=True).split('\n'):
            if len(line)>72 and line[0]==' ':
                lines.append(line[:72]+'&\n     &')
                line = line[72:]
                while len(line)>66:
                    lines.append(line[:66]+'&\n     &')
                    line = line[66:]
                lines.append(line+'\n')
            else: lines.append(line+'\n')
        return ''.join(lines).replace('\n     &\n','\n')

    def format_message(self, kind, message):
        if self.item is not None:
            message = self.reader.format_message(kind, message,
                                                 self.item.span[0], self.item.span[1])
        else:
            return message
        return message

    # def show_message(self, message, stream=sys.stderr):
    #     print >> stream, message
    #     stream.flush()
    #     return

    def error(self, message):
        #message = self.format_message('ERROR', red_text(message)) # KGEN deletion
        message = self.format_message('ERROR', message) # KGEN addition
        logger.error(message)
        # self.show_message(message)
        return

    def warning(self, message):
        #message = self.format_message('WARNING', yellow_text(message)) # KGEN deletion
        message = self.format_message('WARNING', message) # KGEN addition
        logger.warning(message)
        # self.show_message(message)
        return

    def info(self, message):
        message = self.format_message('INFO', message)
        logger.debug(message)
        # self.show_message(message)
        return

    def analyze(self):
        self.warning('nothing analyzed')
        return

    def get_variable(self, name):
        """ Return Variable instance of variable name.
        """
        mth = getattr(self,'get_variable_by_name', self.parent.get_variable)
        return mth(name)

    def get_type(self, name):
        """ Return type declaration using implicit rules
        for name.
        """
        mth = getattr(self,'get_type_by_name', self.parent.get_type)
        return mth(name)

    def get_type_decl(self, kind):
        mth = getattr(self,'get_type_decl_by_kind', self.parent.get_type_decl)
        return mth(kind)

    def get_provides(self):
        """ Returns dictonary containing statements that block provides or None when N/A.
        """
        return

    # start of KGEN

    def remove_label(self, line):
        # remove label
        _label_re = re.compile(r'\s*(?P<label>\d+)\s*(\b|(?=&)|\Z)',re.I)
        m = _label_re.match(line)
        if m:
            label = int(m.group('label'))
            line = line[m.end():]

        return line

    def remove_construct_name(self, line):
        # remove construct name
        _construct_name_re = re.compile(r'\s*(?P<name>\w+)\s*:\s*(\b|(?=&)|\Z)',re.I)
        m = _construct_name_re.match(line)
        if m:
            name = m.group('name')
            line = line[m.end():].lstrip()

        return line

    def tokgen(self):
        raise ProgramException('%s should implement tokgen().'%self.__class__)
#        from statements import Comment, Where
#
#        if isinstance(self, Comment):
#            return self.tofortran().lstrip()
#        # Temporary fix
#        elif isinstance(self, Where):
#            return self.item.line.lstrip()
#        else:
#            return self.item.apply_map(self.tofortran().lstrip())

    def ancestors(self, include_beginsource=False):
        from fortlab.resolver.block_statements import BeginSource, HasUseStmt, Type

        anc = []

        parent = self.parent
        while not isinstance(parent, BeginSource):
            if isinstance(parent, HasUseStmt) or parent.__class__ in [Type]:
                anc.append(parent)
            parent = parent.parent

        if include_beginsource:
            anc.append(parent)

        anc.reverse()
        return anc
 
    def parse_f2003(self):
        from fortlab.resolver.block_statements import BeginSource, SubProgramStatement
        from fortlab.resolver.statements import Continue

        if not hasattr(self, 'f2003'):
            if hasattr(self, 'f2003_class'):
                if hasattr(self.item, 'line') and self.item.line:
                    line = self.tokgen()
                    is_escape = False
                    cstr = None
                    epos = len(line)
                    for i, c in enumerate(line):
                        if is_escape:
                            is_escape = False
                            continue
                        if c=='\\': is_escape = True
                        elif c=='!':
                            if cstr is None:
                                epos = i
                                break
                        elif c=="'" or c=='"':
                            if cstr is None: cstr=c
                            elif cstr==c: cstr = None

                    line = line[:epos]

                    line = self.remove_label(line)
                    line = self.remove_construct_name(line)
#                    # remove label
#                    _label_re = re.compile(r'\s*(?P<label>\d+)\s*(\b|(?=&)|\Z)',re.I)
#                    m = _label_re.match(line)
#                    if m:
#                        label = int(m.group('label'))
#                        line = line[m.end():]
#                    # remove construct name
#                    _construct_name_re = re.compile(r'\s*(?P<name>\w+)\s*:\s*(\b|(?=&)|\Z)',re.I)
#                    m = _construct_name_re.match(line)
#                    if m:
#                        name = m.group('name')
#                        line = line[m.end():].lstrip()
                    self.f2003 = self.f2003_class(line)
                elif hasattr(self.item, 'comment'):
                    #self.f2003 = self.f2003_class(self.item.comment)
                    self.f2003 = self.f2003_class()
                elif self.__class__ in [ BeginSource ]:
                    self.f2003 = self.f2003_class()
                else:
                    raise Exception('Not either line or comment: ', str(self), str(self.__class__))

                if self.f2003 is None:
                    raise Exception('None f2003 attribute: ', str(self))

                self.f2003.stmtpair = self
                traverse(self.f2003, self.set_parent, None)
            else:
                raise ProgramException('Class %s does not have f2003_class attribute.' % self.__class__)

    def set_parent(self, node, bag, depth):
        from fortlab.resolver import Fortran2003

        if hasattr(node, 'item') and node.item and isinstance(node.item, Fortran2003.Base):
            node.item.parent = node

        if node and hasattr(node, 'items') and node.items:
            for item in node.items:
                if isinstance(item, list) or isinstance(item, tuple):
                    for subitem in item:
                        if subitem and not isinstance(subitem, str) and \
                            not isinstance(item, list) and not isinstance(item, tuple):
                            subitem.parent = node
                            self.set_parent(subitem, bag, depth+1)
                elif item and  not isinstance(item, str):
                    item.parent = node
                    self.set_parent(item, bag, depth+1)

    def expr_by_name(self, name, node=None):
        from fortlab.resolver import Fortran2003

        if not name:
            raise ProgramException("Callsite name is not found. Please check if callsite is specified correctly.")

        expr = None
        if isinstance(name, str):
            name = KGName(name) 

        if isinstance(node, Fortran2003.Name):
            if name.firstpartname().lower()==node.string.lower():
                if len(name.namelist)>1:
                    ancnames = [ a.name.lower() for a in self.ancestors() ]
                    lennl = len(name.namelist[:-1])
                    if ancnames[-1*lennl:]==name.namelist[:-1]:
                        expr = node.parent
                else:
                    expr = node.parent
        elif isinstance(node, list) or isinstance(node, tuple):
            for item in node:
                if isinstance(item, Fortran2003.Name):
                    if name.firstpartname().lower()==item.string.lower():
                        if len(name.namelist)>1:
                            ancnames = [ a.name.lower() for a in self.ancestors() ]
                            lennl = len(name.namelist[:-1])
                            if ancnames[-1*lennl:]==name.namelist[:-1]:
                                expr = item.parent
                        else:
                            expr = item.parent
                else:
                    expr = self.expr_by_name(name, item)
                    if expr is not None: break
        else:
            if node and hasattr(node, 'items') and node.items:
                for item in node.items:
                    expr = self.expr_by_name(name, item)
                    if expr is not None: break

        return expr

    def can_resolve(self, request):
        from fortlab.resolver.typedecl_statements import TypeDeclarationStatement
        from fortlab.resolver.block_statements import SubProgramStatement

        if request is None: return False

        # check if request is maded by this stmt itself
        if self is request.originator:
            if isinstance(self, TypeDeclarationStatement):
                if request.uname.firstpartname() in [ n.split('=')[0].strip() for n in self.entity_decls ]:
                    return True
            else:
                return False

        # check if name is matched and self is in resolver classes
        if hasattr(self, 'name') and self.name and \
            self.name.lower()==request.uname.firstpartname() and \
            any( isinstance(self, resolver) for resolver in request.resolvers ):
            return True

        # check if name is one of dummy args in Subprogram
        if isinstance(self, SubProgramStatement) and request.uname.firstpartname() in self.args:
            return True

        return False

    # save names that this self resolved
    def add_geninfo(self, uname, request):

        if uname is None or request is None: return

        # Statement
        if not hasattr(self, 'geninfo'):
            self.geninfo = OrderedDict()
        if request.gentype not in self.geninfo:
            self.geninfo[request.gentype] = []
        if (uname, request) not in self.geninfo[request.gentype]:
            self.geninfo[request.gentype].append((uname, request))

        # EndStatement
        if isinstance(self, BeginStatement) and isinstance(self.content[-1], EndStatement):
            if not hasattr(self.content[-1], 'geninfo'):
                self.content[-1].geninfo = OrderedDict()
            if request.gentype not in self.content[-1].geninfo:
                self.content[-1].geninfo[request.gentype] = []

        # Ancestors
        for anc in self.ancestors(include_beginsource=True):
            if not hasattr(anc, 'geninfo'):
                anc.geninfo = OrderedDict()
            if request.gentype not in anc.geninfo:
                anc.geninfo[request.gentype] = []
            if isinstance(anc, BeginStatement) and isinstance(anc.content[-1], EndStatement):
                if not hasattr(anc.content[-1], 'geninfo'):
                    anc.content[-1].geninfo = OrderedDict()
                if request.gentype not in anc.content[-1].geninfo:
                    anc.content[-1].geninfo[request.gentype] = []


    def check_spec_stmts(self, uname, request, config):
        # the last resolver
        res_stmt = request.res_stmts[-1]
        if not hasattr(res_stmt, 'parent') or not hasattr(res_stmt.parent, 'spec_stmts'):
            return

        for spec_stmt in res_stmt.parent.spec_stmts:
            spec_stmt.resolve_uname(uname, request, config)


    def get_res_stmts(self, uname):
        if not hasattr(self, 'unknowns'):
            return None

        if isinstance(uname, str): strname = uname
        elif isinstance(uname, KGName): strname = uname.firstpartname()

        for kgname, res in self.unknowns.items():
            if res is None:
                return None
            if kgname.firstpartname()==strname:
                return res.res_stmts
        return None

    def resolve(self, request, config):
        from fortlab.resolver.kgparse import ResState
        from fortlab.resolver.block_statements import BeginSource
        from fortlab.resolver.typedecl_statements import TypeDeclarationStatement
        from fortlab.resolver.kgsearch import f2003_search_unknowns
        from fortlab.resolver.api import walk
        from fortlab.resolver.block_statements import SubProgramStatement

        if request is None: return
        logger.debug('%s is being resolved'%request.uname.firstpartname())

        # already resolved
        if request.state == ResState.RESOLVED:
            logger.debug('%s is already resolved'%request.uname.firstpartname())
            return

        # check if this stmt can resolve the request
        if self.can_resolve(request):
            if isinstance(self, TypeDeclarationStatement):
                typedecl_stmt  = self
                logger.debug('The request is being resolved by a variable')
                request.res_stmts.append(typedecl_stmt)
                request.state = ResState.RESOLVED
                typedecl_stmt.add_geninfo(request.uname, request)
                self.check_spec_stmts(request.uname, request, config)
                logger.debug('%s is resolved'%request.uname.firstpartname())

                if not hasattr(typedecl_stmt, 'unknowns'):
                    f2003_search_unknowns(typedecl_stmt, typedecl_stmt.f2003, config)
                if hasattr(typedecl_stmt, 'unknowns'):
                    for unk, req in typedecl_stmt.unknowns.items():
                        if req.state != ResState.RESOLVED:
                            typedecl_stmt.resolve(req, config)
            elif isinstance(self, SubProgramStatement):
                if not hasattr(self, 'implicit_rule_resolvers'): self.implicit_rule_resolvers = []
                irres = ImplicitRule_Resolver(request.uname.firstpartname())
                irres.add_geninfo(request.uname, request)
                self.implicit_rule_resolvers.append(irres)
                request.res_stmts.append(irres)
                request.state = ResState.RESOLVED
                logger.debug('%s is resolved by Subprogram %s using implicit rules'%(request.uname.firstpartname(), self.name))
            else:
                logger.debug('%s can be resolved'%request.uname.firstpartname())
                raise Exception('This request should have been resolved by the stmt shown below.\n%s'%self.tokgen())

        # ask parent for resolution
        if request.state != ResState.RESOLVED:
            logger.debug('%s is not resolved locally and the request is being defered to parent'%request.uname.firstpartname())
            if hasattr(self, 'parent') and not isinstance(self.parent, BeginSource):
                self.parent.resolve(request, config)


        # if not resolved,
        if request.state != ResState.RESOLVED:
            if self is request.originator:
                # check if program units can resolve the request
                for filepath, units in config["program_units"].items():
                    for unit in units:
                        if any( isinstance(unit, resolver) for resolver in request.resolvers) and \
                            hasattr(unit, 'name') and request.uname.firstpartname()==unit.name:
                            logger.debug('The request is being resolved by a program unit')
                            if unit not in request.originator.ancestors():
                                request.res_stmts.append(unit)
                                request.state = ResState.RESOLVED
                                unit.add_geninfo(request.uname, request)
                                self.check_spec_stmts(request.uname, request, config)
                                logger.debug('%s is resolved'%request.uname.firstpartname())
                                for _stmt, _depth in walk(unit, -1):
                                    if not hasattr(_stmt, 'unknowns'):
                                        f2003_search_unknowns(_stmt, _stmt.f2003, config)
                                    if hasattr(_stmt, 'unknowns'):
                                        for unk, req in _stmt.unknowns.items():
                                            if req.state != ResState.RESOLVED:
                                                _stmt.resolve(req, config) 

                                # if newly found program unit is not in srcfiles
                                if not unit in config["srcfiles"][self.top.reader.id][2]:
                                    config["srcfiles"][self.top.reader.id][2].append(unit)
                    if request.state==ResState.RESOLVED:
                        break

                # check if intrinsic procedure can resolve
                if request.state != ResState.RESOLVED:
                    if request.uname.firstpartname() in Intrinsic_Procedures:
                        excepts = config["search"]['except']
                        if  config["search"]['skip_intrinsic'] and not is_except(request.uname.firstpartname(), self, excepts) or \
                            not config["search"]['skip_intrinsic'] and is_except(request.uname.firstpartname(), self, excepts):
                            ipres = IntrinsicProcedure_Resolver(request.uname.firstpartname())
                            ipres.add_geninfo(request.uname, request)
                            request.res_stmts.append(ipres)
                            request.state = ResState.RESOLVED
                            logger.debug('%s is resolved using intrinsic procedure'%request.uname.firstpartname())

                # tries to apply implicit rules
                if request.state != ResState.RESOLVED:
                    try:
                        mytype = self.get_type(request.uname.firstpartname())
                        for resolver in request.resolvers:
                            if isinstance(mytype, resolver):
                                anc = self.ancestors()
                                if not hasattr(anc[-1], 'implicit_rule_resolvers'): anc[-1].implicit_rule_resolvers = []
                                irres = ImplicitRule_Resolver(request.uname.firstpartname())
                                irres.add_geninfo(request.uname, request)
                                anc[-1].implicit_rule_resolvers.append(irres)
                                request.res_stmts.append(irres)
                                request.state = ResState.RESOLVED
                                logger.debug('%s is resolved using implicit rules'%request.uname.firstpartname())
                                break
                        #import pdb; pdb.set_trace()
                        # TODO: check against implicit_rules of parent
                        # TODO: mark implicit resolution in ResState
                    except:
                        #import pdb; pdb.set_trace()
                        pass

                # if still not resolved, mark it
                if request.state != ResState.RESOLVED:
                    logger.debug('%s is not resolved' % request.uname.firstpartname())
                    # mark it
        else:
            logger.debug('%s is resolved' % request.uname.firstpartname())
 
    # end of KGEN

class BeginStatement(Statement):
    """[ construct_name : ] <blocktype> [ <name> ]

    BeginStatement instances have additional attributes:
      name
      blocktype

    Block instance has attributes:
      content - list of Line or Statement instances
      name    - name of the block, unnamed blocks are named
                with the line label
      construct_name - name of a construct
      parent  - Block or FortranParser instance
      item    - Line instance containing the block start statement
      get_item, put_item - methods to retrive/submit Line instances
                from/to Fortran reader.
      isvalid - boolean, when False, the Block instance will be ignored.

      stmt_cls, end_stmt_cls

    """
    _repr_attr_names = ['blocktype','name','construct_name'] + Statement._repr_attr_names
    def __init__(self, parent, item=None):

        self.typedecl_resolvers = [] # KGen addition for variable list app

        self.content = []
        self.get_item = parent.get_item # get line function
        self.put_item = parent.put_item # put line function
        if not hasattr(self, 'blocktype'):
            self.blocktype = self.__class__.__name__.lower()
        if not hasattr(self, 'name'):
            # process_item may change this
            self.name = '__'+self.blocktype.upper()+'__'
        self.construct_name = getattr(item,'name',None)
        Statement.__init__(self, parent, item)
        return

    # start of KGEN
    def check_access(self, req):

        ancs = req.originator.ancestors()
        if self in ancs: return True

        if not hasattr(self, 'check_private'): return False

        uname = req.uname.firstpartname()
        if self.check_private(uname): return False
        else: return True

    def resolve(self, request, config):
        from fortlab.resolver.kgparse import ResState
        from fortlab.resolver.kgsearch import f2003_search_unknowns
        from fortlab.kgutils import pack_exnamepath
        from fortlab.resolver.block_statements import HasUseStmt, Type, TypeDecl, Function, Subroutine, Interface, Associate
        from fortlab.resolver.typedecl_statements import TypeDeclarationStatement
        from fortlab.resolver.statements import External, Use, GenericBinding, SpecificBinding, Call
        from fortlab.resolver.api import walk
        from fortlab.resolver.block_statements import SubProgramStatement
        from fortlab.resolver import Fortran2003

        if request is None: return

        logger.debug('%s is being resolved'%request.uname.firstpartname())

        # if resolved, return
        if request.state == ResState.RESOLVED:
            logger.debug('%s is already resolved'%request.uname.firstpartname())
            return

        # check if stmt is either subprogram, module, interface
        if isinstance(self, HasUseStmt):

            # NOTE: check if the resolver is in the file or in other file
            # check if internal a subprogram can resolve
            if request.state != ResState.RESOLVED:
                subp = None
                if hasattr(self.a, 'internal_subprogram') and request.uname.firstpartname() in self.a.internal_subprogram.keys():
                    subp = self.a.internal_subprogram[request.uname.firstpartname()]
                elif hasattr(self.a, 'module_subprogram') and request.uname.firstpartname() in self.a.module_subprogram.keys():
                    subp = self.a.module_subprogram[request.uname.firstpartname()]
                elif hasattr(self.a, 'module_interface'):
                    for if_stmt in self.a.module_interface:
                        if request.uname.firstpartname()==if_stmt.name:
                            subp = if_stmt
                            break

                if subp and any( isinstance(subp, resolver) for resolver in request.resolvers ):
                    logger.debug('The request is being resolved by a subprogram or interface')
                    request.res_stmts.append(subp)
                    request.state = ResState.RESOLVED
                    subp.add_geninfo(request.uname, request)
                    self.check_spec_stmts(request.uname, request, config)
                    logger.debug('%s is resolved'%request.uname.firstpartname())

                    if subp not in request.originator.ancestors():
                        for _stmt, _depth in walk(subp, -1):
                            if not hasattr(_stmt, 'unknowns'):
                                f2003_search_unknowns(_stmt, _stmt.f2003, config)
                            if hasattr(_stmt, 'unknowns'):
                                for unk, req in _stmt.unknowns.items():
                                    if req.state != ResState.RESOLVED:
                                        _stmt.resolve(req, config) 

            # check if self is a subprogram and it can resolve
            if request.state != ResState.RESOLVED:
                subp = None
                if isinstance(self, SubProgramStatement) and request.uname.firstpartname()==self.name:
                    subp = self

                if subp and any( isinstance(subp, resolver) for resolver in request.resolvers ):
                    logger.debug('The request is being resolved by a subprogram or interface')
                    request.res_stmts.append(subp)
                    request.state = ResState.RESOLVED
                    subp.add_geninfo(request.uname, request)
                    self.check_spec_stmts(request.uname, request, config)
                    logger.debug('%s is resolved'%request.uname.firstpartname())

                    if subp not in request.originator.ancestors():
                        for _stmt, _depth in walk(subp, -1):
                            if not hasattr(_stmt, 'unknowns'):
                                f2003_search_unknowns(_stmt, _stmt.f2003, config)
                            if hasattr(_stmt, 'unknowns'):
                                for unk, req in _stmt.unknowns.items():
                                    if req.state != ResState.RESOLVED:
                                        _stmt.resolve(req, config) 

# TODO: With this, KGen fails to resolve for variables defined in different module
#                if isinstance(self, SubProgramStatement) and any( request.uname.firstpartname() == arg for arg in self.args ):
#                    if TypeDeclarationStatement in request.resolvers:
#                        logger.debug('The request is being resolved by dummy argument of a subprogram')
#                        request.res_stmts.append(subp)
#                        request.state = ResState.RESOLVED
#                        self.add_geninfo(request.uname, request)
#                        #self.check_spec_stmts(request.uname, request)
#                        logger.debug('%s is resolved'%request.uname.firstpartname())

            # check if subprogram stmt in Interface block can resolve
            if request.state != ResState.RESOLVED and isinstance(request.originator, SpecificBinding) and \
                hasattr(self.a, 'module_interface') and Interface in request.resolvers:
                for if_obj in self.a.module_interface:
                    if if_obj.isabstract:
                        for item in if_obj.content:
                            if item.__class__ in [ Function, Subroutine ] and request.uname.firstpartname()==item.name:
                                logger.debug('The request is being resolved by a Subprogram in abstract interface')
                                request.res_stmts.append(item)
                                request.state = ResState.RESOLVED
                                item.add_geninfo(request.uname, request)
                                self.check_spec_stmts(request.uname, request, config)
                                logger.debug('%s is resolved'%request.uname.firstpartname())
                                for _stmt, _depth in walk(item, -1):
                                    if not hasattr(_stmt, 'unknowns'):
                                        f2003_search_unknowns(_stmt, _stmt.f2003, config)
                                    if hasattr(_stmt, 'unknowns'):
                                        for unk, req in _stmt.unknowns.items():
                                            if req.state != ResState.RESOLVED:
                                                _stmt.resolve(req, config) 

            # check if a type can resolve
            if request.state != ResState.RESOLVED and hasattr(self.a, 'type_decls') and \
                request.uname.firstpartname() in self.a.type_decls.keys():
                type_stmt  = self.a.type_decls[request.uname.firstpartname()]
                if any( isinstance(type_stmt, resolver) for resolver in request.resolvers ):
                    logger.debug('The request is being resolved by a typedecl')
                    request.res_stmts.append(type_stmt)
                    request.state = ResState.RESOLVED
                    type_stmt.add_geninfo(request.uname, request)
                    self.check_spec_stmts(request.uname, request, config)
                    logger.debug('%s is resolved'%request.uname.firstpartname())

                    for _stmt, _depth in walk(type_stmt, -1):
                        if not hasattr(_stmt, 'unknowns'):
                            f2003_search_unknowns(_stmt, _stmt.f2003, config)
                        if hasattr(_stmt, 'unknowns'):
                            for unk, req in _stmt.unknowns.items():
                                if req.state != ResState.RESOLVED:
                                    _stmt.resolve(req, config) 

            # check if a enum can resolve
            if request.state != ResState.RESOLVED and hasattr(self.a, 'enum_decls') and \
                request.uname.firstpartname() in self.a.enum_decls.keys():
                enum_stmt, enumerator_stmt  = self.a.enum_decls[request.uname.firstpartname()]
                if any( isinstance(enumerator_stmt, resolver) for resolver in request.resolvers ):
                    logger.debug('The request is being resolved by a enumdecl')
                    request.res_stmts.append(enum_stmt)
                    request.state = ResState.RESOLVED
                    enum_stmt.add_geninfo(request.uname, request)
                    self.check_spec_stmts(request.uname, request, config)
                    logger.debug('%s is resolved'%request.uname.firstpartname())

                    for _stmt, _depth in walk(enum_stmt, -1):
                        if not hasattr(_stmt, 'unknowns'):
                            f2003_search_unknowns(_stmt, _stmt.f2003, config)
                        if hasattr(_stmt, 'unknowns'):
                            for unk, req in _stmt.unknowns.items():
                                if req.state != ResState.RESOLVED:
                                    _stmt.resolve(req, config) 

            # check if a module variable can resolve
            # NOTE: check if the resolver is in the file or in other file
            if request.state != ResState.RESOLVED and hasattr(self.a, 'variables') and \
                request.uname.firstpartname() in self.a.variables.keys():
                #self.a.variables[request.uname.firstpartname()].is_public():

                typedecl_stmt  = self.a.variables[request.uname.firstpartname()].parent

                #if ((isinstance(res_stmt, TypeDeclarationStatement) and request.uname.firstpartname() in res_stmt.entity_decls) or \
                #    res_stmt is not request.originator) and any( isinstance(res_stmt, resolver) for resolver in request.resolvers ):
                if any( isinstance(typedecl_stmt, resolver) for resolver in request.resolvers ):
                    logger.debug('The request is being resolved by a variable')
                    request.res_stmts.append(typedecl_stmt)
                    request.state = ResState.RESOLVED
                    typedecl_stmt.add_geninfo(request.uname, request)
                    self.check_spec_stmts(request.uname, request, config)
                    logger.debug('%s is resolved'%request.uname.firstpartname())

                    if typedecl_stmt not in self.typedecl_resolvers:
                        self.typedecl_resolvers.append(typedecl_stmt)

                    if not hasattr(typedecl_stmt, 'unknowns'):
                        f2003_search_unknowns(typedecl_stmt, typedecl_stmt.f2003, config)
                    if hasattr(typedecl_stmt, 'unknowns'):
                        for unk, req in typedecl_stmt.unknowns.items():
                            if req.state != ResState.RESOLVED:
                                typedecl_stmt.resolve(req, config) 

            # check if an interface block in a module can resolve
            if request.state != ResState.RESOLVED and isinstance(request.originator, Call) and \
                hasattr(self.a, 'module_interface') and Interface in request.resolvers:
                for if_obj in self.a.module_interface:
                    for item in if_obj.content:
                        if item.__class__ in [ Function, Subroutine ] and request.uname.firstpartname()==item.name:
                            logger.info('The request is being resolved by a Subprogram in an interface')
                            request.res_stmts.append(item)
                            request.state = ResState.RESOLVED
                            item.add_geninfo(request.uname, request)
                            self.check_spec_stmts(request.uname, request, config)
                            logger.info('%s is resolved'%request.uname.firstpartname())
                            for _stmt, _depth in walk(item, -1):
                                if not hasattr(_stmt, 'unknowns'):
                                    f2003_search_unknowns(_stmt, _stmt.f2003, config)
                                if hasattr(_stmt, 'unknowns'):
                                    for unk, req in _stmt.unknowns.items():
                                        if req.state != ResState.RESOLVED:
                                            _stmt.resolve(req, config) 

            # check if an interface block in a subprogram can resolve
            if request.state != ResState.RESOLVED and isinstance(request.originator, Call) and \
                hasattr(self.a, 'subprogram_interface') and Interface in request.resolvers:
                for if_obj in self.a.subprogram_interface:
                    for item in if_obj.content:
                        if item.__class__ in [ Function, Subroutine ] and request.uname.firstpartname()==item.name:
                            logger.info('The request is being resolved by a Subprogram in an interface')
                            request.res_stmts.append(item)
                            request.state = ResState.RESOLVED
                            item.add_geninfo(request.uname, request)
                            self.check_spec_stmts(request.uname, request, config)
                            logger.info('%s is resolved'%request.uname.firstpartname())
                            for _stmt, _depth in walk(item, -1):
                                if not hasattr(_stmt, 'unknowns'):
                                    f2003_search_unknowns(_stmt, _stmt.f2003, config)
                                if hasattr(_stmt, 'unknowns'):
                                    for unk, req in _stmt.unknowns.items():
                                        if req.state != ResState.RESOLVED:
                                            _stmt.resolve(req, config) 

            # check if a common statement can resolve
            #if request.state != ResState.RESOLVED and hasattr(self.a, 'common_stmts') and \
            #    len(self.a.common_stmts)>0:
            if request.state != ResState.RESOLVED and hasattr(self.a, 'common_stmts') and \
                request.uname.firstpartname() in self.a.common_stmts.keys():
                common_stmt  = self.a.common_stmts[request.uname.firstpartname()]
                if any( isinstance(common_stmt, resolver) for resolver in request.resolvers ):
                    logger.debug('The request is being resolved by a common stmt')
                    request.res_stmts.append(common_stmt)
                    request.state = ResState.RESOLVED
                    common_stmt.add_geninfo(request.uname, request)
                    self.check_spec_stmts(request.uname, request, config)
                    logger.debug('%s is resolved'%request.uname.firstpartname())

                    if not hasattr(common_stmt, 'unknowns'):
                        f2003_search_unknowns(common_stmt, common_stmt.f2003, config)
                    if hasattr(common_stmt, 'unknowns'):
                        for unk, req in common_stmt.unknowns.items():
                            if req.state != ResState.RESOLVED:
                                common_stmt.resolve(req, config) 

            # handle external
            if request.state != ResState.RESOLVED and isinstance(request.originator, External) and \
                self.__class__ in [ Function, Subroutine ] and request.uname.firstpartname() in self.args:
                external_stmt = self
                request.res_stmts.append(external_stmt)
                request.state = ResState.RESOLVED
                external_stmt.add_geninfo(request.uname, request)
                self.check_spec_stmts(request.uname, request, config)
                logger.debug('%s is resolved'%request.uname.firstpartname())

            # check if use stmt can resolve
            if request.state != ResState.RESOLVED:
                logger.debug('%s is not resolved on this block'%request.uname.firstpartname())
                if hasattr(self, 'use_stmts') and len(self.use_stmts)>0:
                    logger.debug('The request is being deferred to use stmt')
                    uname = request.uname.firstpartname()

                    # first, try with use stmts having only keyword
                    for mod_name, use_stmts in self.use_stmts.items():
                        for use_stmt in use_stmts:
                            if use_stmt.isonly:
                                if uname in use_stmt.norenames and self.check_access(request):
                                    logger.debug('%s is found in norenames'%uname)
                                    use_stmt.resolve(request, config)
                                    if request.state == ResState.RESOLVED:
                                        logger.debug('%s is resolved in norenames'%uname)
                                        request.res_stmts.append(use_stmt)
                                        use_stmt.add_geninfo(request.uname, request)
                                        self.check_spec_stmts(request.uname, request, config)
                                        break

                                rename = [r for r in use_stmt.renames if r[0]==uname]
                                if len(rename)>1:
                                    raise ProgramException('More than one result: %s'%str(rename))
                                elif len(rename)==1 and self.check_access(request):
                                    newname = KGName(pack_exnamepath(request.originator, rename[0][1]), node=use_stmt.f2003, stmt=use_stmt)
                                    request.push_uname(newname)
                                    use_stmt.resolve(request, config)
                                    if request.state == ResState.RESOLVED:
                                        request.pop_uname(reset_uname=True)
                                        request.res_stmts.append(use_stmt)
                                        use_stmt.add_geninfo(request.uname, request)
                                        self.check_spec_stmts(request.uname, request, config)
                                        break
                                    else:
                                        request.pop_uname()
                        if request.state==ResState.RESOLVED: break

                    # and then, try with use stmts not having only keyword
                    # but still has renames or norenames
                    if request.state != ResState.RESOLVED:
                        for mod_name, use_stmts in self.use_stmts.items():
                            for use_stmt in use_stmts:
                                if not use_stmt.isonly:
                                    if uname in use_stmt.norenames and self.check_access(request):
                                        use_stmt.resolve(request, config)
                                        if request.state == ResState.RESOLVED:
                                            request.res_stmts.append(use_stmt)
                                            use_stmt.add_geninfo(request.uname, request)
                                            self.check_spec_stmts(request.uname, request, config)
                                            break

                                    rename = [r for r in use_stmt.renames if r[0]==uname]
                                    if len(rename)>1:
                                        raise ProgramException('More than one result: %s'%str(rename))
                                    elif len(rename)==1 and self.check_access(request):
                                        newname = KGName(pack_exnamepath(request.originator, rename[0][1]), node=use_stmt.f2003, stmt=use_stmt)
                                        request.push_uname(newname)
                                        use_stmt.resolve(request, config)

                                        if request.state == ResState.RESOLVED:
                                            request.pop_uname(reset_uname=True)
                                            request.res_stmts.append(use_stmt)
                                            use_stmt.add_geninfo(request.uname, request)
                                            self.check_spec_stmts(request.uname, request, config)
                                            break
                                        else:
                                            request.pop_uname()
                            if request.state==ResState.RESOLVED: break

#                    # and then, try with use stmts that has an analyzed-module
#                    if request.state != ResState.RESOLVED:
#                        for mod_name, use_stmts in self.use_stmts.items():
#                            for use_stmt in use_stmts:
#                                if use_stmt.module:
#                                    if uname in use_stmt.module.a.module_provides:
#                                        use_stmt.resolve(request)
#                                        if request.state == ResState.RESOLVED:
#                                            request.res_stmt = use_stmt
#                                            request.res_stmt.add_geninfo(request.uname, request)
#                                            self.check_spec_stmts(request.uname, request)
#                                            break
#
#                                    rename = [r for r in use_stmt.renames if r[0]==uname]
#                                    if len(rename)>1:
#                                        raise ProgramException('More than one result: %s'%str(rename))
#                                    elif len(rename)==1 and uname in use_stmt.module.a.module_provides:
#                                        request.uname.set_name(rename[0][1])
#                                        use_stmt.resolve(request)
#                                        request.uname.reset_name()
#                                        if request.state == ResState.RESOLVED:
#                                            request.res_stmt = use_stmt
#                                            request.res_stmt.add_geninfo(request.uname, request)
#                                            self.check_spec_stmts(request.uname, request)
#                                            break
#                            if request.state==ResState.RESOLVED: break

                    # and then, try with use stmts not having only keyword
                    # and not has renames or norenames
                    if request.state != ResState.RESOLVED:
                        for mod_name, use_stmts in self.use_stmts.items():
                            for use_stmt in use_stmts:
                                if not use_stmt.isonly:
                                    if len(use_stmt.norenames)==0 and len(use_stmt.renames)==0 and self.check_access(request):
                                        use_stmt.resolve(request, config)
                                        if request.state == ResState.RESOLVED:
                                            request.res_stmts.append(use_stmt)
                                            use_stmt.add_geninfo(request.uname, request)
                                            self.check_spec_stmts(request.uname, request, config)
                                            break
                            if request.state==ResState.RESOLVED: break

        elif self.__class__ in [ Type, TypeDecl ]:
            # check if a type can resolve
            if request.state != ResState.RESOLVED and request.uname.firstpartname()==self.name:
                type_stmt  = self
                if any( isinstance(type_stmt, resolver) for resolver in request.resolvers ):
                    logger.debug('The request is being resolved by a type')
                    request.res_stmts.append(type_stmt)
                    request.state = ResState.RESOLVED
                    type_stmt.add_geninfo(request.uname, request)
                    self.check_spec_stmts(request.uname, request, config)
                    logger.debug('%s is resolved'%request.uname.firstpartname())

                    #if self not in request.originator.ancestors():
                    if self is request.originator.parent:
                        pass
                    else:
                        for _stmt, _depth in walk(type_stmt, -1):
                            if not hasattr(_stmt, 'unknowns'):
                                f2003_search_unknowns(_stmt, _stmt.f2003, config)
                            if hasattr(_stmt, 'unknowns'):
                                for unk, req in _stmt.unknowns.items():
                                    if req.state != ResState.RESOLVED:
                                        _stmt.resolve(req, config)
            elif request.state != ResState.RESOLVED and isinstance(request.originator, GenericBinding):
                for child in self.content:
                    if isinstance(child, SpecificBinding):
                        if request.uname.firstpartname()==child.name:
                            logger.debug('The request is being resolved by a specific binding')
                            request.res_stmts.append(child)
                            request.state = ResState.RESOLVED
                            child.add_geninfo(request.uname, request)
                            child.check_spec_stmts(request.uname, request, config)
                            logger.debug('%s is resolved'%request.uname.firstpartname())

                            #if self not in request.originator.ancestors():
                            for _stmt, _depth in walk(child, -1):
                                if not hasattr(_stmt, 'unknowns'):
                                    f2003_search_unknowns(_stmt, _stmt.f2003, config)
                                if hasattr(_stmt, 'unknowns'):
                                    for unk, req in _stmt.unknowns.items():
                                        if req.state != ResState.RESOLVED:
                                            _stmt.resolve(req, config)

        elif isinstance(self, Associate):
            def get_name(node, bag, depth):
                if isinstance(node, Fortran2003.Association):
                    bag[node.items[0].string] = node.items[2]

            # check if associate construct can resolve
            assocnames = {}
            traverse(self.f2003, get_name, assocnames)

            if request.state != ResState.RESOLVED and request.uname.firstpartname() in assocnames:
                assoc_stmt  = self
                if any( isinstance(assoc_stmt, resolver) for resolver in request.resolvers ):
                    logger.debug('The request is being resolved by a association construct')
                    request.res_stmts.append(assoc_stmt)
                    request.state = ResState.RESOLVED
                    assoc_stmt.add_geninfo(request.uname, request)
                    logger.debug('%s is resolved'%request.uname.firstpartname())

                    #if self not in request.originator.ancestors():
                    if not hasattr(assoc_stmt, 'unknowns'):
                        f2003_search_unknowns(assoc_stmt, assocnames[request.uname.firstpartname()], config)

                    if hasattr(assoc_stmt, 'unknowns'):
                        if not hasattr(assoc_stmt, 'assoc_map'):
                            assoc_stmt.assoc_map = {}
                        for unk, req in assoc_stmt.unknowns.items():

                            if unk in assoc_stmt.assoc_map:
                                if request.uname not in assoc_stmt.assoc_map[unk]:
                                    assoc_stmt.assoc_map[unk].append(request.uname)
                            else:
                                assoc_stmt.assoc_map[unk] = [ request.uname ]

                            if req.state != ResState.RESOLVED:
                                assoc_stmt.resolve(req, config)

        # defer to super
        if request.state != ResState.RESOLVED:
            super(BeginStatement, self).resolve(request, config)

    def tokgen(self):
        construct_name = self.construct_name
        construct_name = construct_name + ': ' if construct_name else ''
        if hasattr(self, 'item') and hasattr(self.item, 'apply_map'):
            return construct_name + self.item.apply_map(self.tostr())
        else:
            return construct_name + self.tostr()
    # end of KGEN

    def tostr(self):
        return self.blocktype.upper() + ' '+ self.name

    def tofortran(self, isfix=None):
        construct_name = self.construct_name
        construct_name = construct_name + ': ' if construct_name else ''
        l=[self.get_indent_tab(isfix=isfix) + construct_name + self.tostr()]
        for c in self.content:
            l.append(c.tofortran(isfix=isfix))
        return '\n'.join(l)

    def torepr(self, depth=-1, incrtab=''):
        tab = incrtab + self.get_indent_tab()
        ttab = tab + '  '
        l=[Statement.torepr(self, depth=depth,incrtab=incrtab)]
        if depth==0 or not self.content:
            return '\n'.join(l)
        l.append(ttab+'content:')
        for c in self.content:
            if isinstance(c,EndStatement):
                l.append(c.torepr(depth-1,incrtab))
            else:
                l.append(c.torepr(depth-1,incrtab + '  '))
        return '\n'.join(l)

    def process_item(self):
        """ Process the line
        """
        item = self.item
        if item is None: return
        self.fill()
        return

    def fill(self, end_flag = False):
        """
        Fills blocks content until the end of block statement.
        """

        mode = self.reader.mode
        class_list = self.get_classes()
        self.classes = [cls for cls in class_list if mode in cls.modes]
        self.pyf_classes = [cls for cls in class_list if 'pyf' in cls.modes]

        item = self.get_item()
        while item is not None:
            if isinstance(item, Line):
                if self.process_subitem(item):
                    end_flag = True
                    break
            elif isinstance(item, Comment):
                #KGEN self.content.append(classes.Comment(self, item))
                from fortlab.resolver.statements import Comment as newComment 
                self.content.append(newComment(self, item))
            else:
                raise NotImplementedError(str(item))
            item = self.get_item()

        if not end_flag:
            self.warning('failed to find the end of block')
        return

    def process_subitem(self, item):
        """
        Check is item is blocks start statement, if it is, read the block.

        Return True to stop adding items to given block.
        """
        line = item.get_line()

        # First check for the end of block
        cls = self.end_stmt_cls
        if cls.match(line):
            stmt = cls(self, item)
            if stmt.isvalid:
                self.content.append(stmt)
                return True

        if item.is_f2py_directive:
            classes = self.pyf_classes
        else:
            classes = self.classes

        # Look for statement match
        for cls in classes:
            if cls.match(line):
                stmt = cls(self, item)
                if stmt.isvalid:
                    if not stmt.ignore:
                        self.content.append(stmt)
                    return False
                # item may be cloned that changes the items line:
                line = item.get_line()

        # Check if f77 code contains inline comments or other f90
        # constructs that got undetected by get_source_info.
        if item.reader.isf77:
            i = line.find('!')
            if i != -1:
                message = item.reader.format_message(\
                        'WARNING',
                        'no parse pattern found for "%s" in %r block,'\
                        ' trying to remove inline comment (not in Fortran 77).'\
                        % (item.get_line(),self.__class__.__name__),
                        item.span[0], item.span[1])
                # .. but at the expense of loosing the comment.
                logger.warning(message)
                # self.show_message(message)
                if line[:i]:
                    newitem = item.copy(line[:i].rstrip())
                    return self.process_subitem(newitem)
                else:
                    return True

            # try fix statement classes
            f77_classes = self.classes
            classes = []
            for cls in self.get_classes():
                if 'f77' in cls.modes and cls not in f77_classes:
                    classes.append(cls)
            if classes:
                message = item.reader.format_message(\
                        'WARNING',
                        'no parse pattern found for "%s" in %r block'\
                        ' maybe due to strict f77 mode.'\
                        ' Trying f90 fix mode patterns..'\
                        % (item.get_line(),self.__class__.__name__),
                        item.span[0], item.span[1])
                logger.warning(message)
                # self.show_message(message)

                item.reader.set_mode(False, False)
                self.classes = classes

                r = BeginStatement.process_subitem(self, item)
                if r is None:
                    # restore f77 fix mode
                    self.classes = f77_classes
                    item.reader.set_mode(False, True)
                else:
                    message = item.reader.format_message(\
                        'INFORMATION',
                        'The f90 fix mode resolved the parse pattern issue.'\
                        ' Setting reader to f90 fix mode.',
                        item.span[0], item.span[1])
                    logger.debug(message)
                    # self.show_message(message)
                    # set f90 fix mode
                    self.classes = f77_classes + classes
                    self.reader.set_mode(False, False)
                return r

        self.handle_unknown_item(item)
        return

    def handle_unknown_item(self, item):
        message = item.reader.format_message(\
                        'WARNING',
                        'no parse pattern found for "%s" in %r block.'\
                        % (item.get_line(),self.__class__.__name__),
                        item.span[0], item.span[1])
        logger.warning(message)
        # self.show_message(message)
        self.content.append(item)
        #sys.exit()
        return

    def analyze(self):
        for stmt in self.content:
            stmt.analyze()
        return

class EndStatement(Statement):
    """
    END [<blocktype> [<name>]]

    EndStatement instances have additional attributes:
      name
      blocktype
    """
    _repr_attr_names = ['blocktype','name'] + Statement._repr_attr_names

    def __init__(self, parent, item):
        if not hasattr(self, 'blocktype'):
            self.blocktype = self.__class__.__name__.lower()[3:]
        Statement.__init__(self, parent, item)

    def process_item(self):
        item = self.item
        line = item.get_line().replace(' ','')[3:]
        line = item.apply_map(line)
        blocktype = self.blocktype
        if line.lower().startswith(blocktype):
            line = line[len(blocktype):].strip()
        else:
            if line:
                # not the end of expected block
                line = ''
                self.isvalid = False
        if self.parent.construct_name:
            name = self.parent.construct_name
        else:
            name = self.parent.name

        if line:
            # line variable is already cast to lower case so would fail if any
            # upper case letters exist in the label. Also, fortran is case
            # insensitive anyway so we should assume labels may have a
            # different case and therefore cast both to the same case in our
            # equivalence test.
            #if line.lower()!=name.lower(): # KGEN deletion
            if line.lower()!=name.replace(' ','').lower(): # KGen addition
                self.warning(\
                    'expected the end of %r block but got the end of %r, skipping.'\
                    % (name, line))
                self.isvalid = False
        # start of KGEN addition
            else:
                self.name = line.lower()
        else:
            self.name = None
        # end of KGEN addition
        #self.name = name # KGEN deletion

    def analyze(self):
        return

    def get_indent_tab(self,deindent=False,isfix=None):
        return Statement.get_indent_tab(self, deindent=True,isfix=isfix)

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'END %s %s'\
               % (self.blocktype.upper(),self.name or '')

    # start of KGEN addition
    def tokgen(self):
        return 'END %s %s' % (self.blocktype.upper(),self.name or '')
    # end of KGEN addition



from microapp import App

class FortranStateGenerator(App):

    _name_ = "stategen"
    _version_ = "0.1.0"

    def __init__(self, mgr):

        self.add_argument("analysis", help="analysis object")
        self.add_argument("--outdir", help="output directory")

        self.register_forward("statedir", help="state generation code directory")
    
    def perform(self, args):

        self.config = args.analysis["_"]

        # create directory if needed
        args.outdir = args.outdir["_"] if args.outdir else os.getcwd()

        if not os.path.exists(args.outdir):
            os.makedirs(args.outdir)

        import pdb; pdb.set_trace()
        state_realpath = os.path.realpath(os.path.join(args.outdir, "state"))

        self.add_forward(statedir=state_realpath)





# uncompyle6 version 3.7.2
# Python bytecode 2.7 (62211)
# Decompiled from: Python 3.6.9 (default, Apr 18 2020, 01:56:04)
# [GCC 8.4.0]
# Embedded file name: /home/grnydawn/repos/github/fortlab/fortlab/kernel/kernelgen.py
# Compiled at: 2020-07-13 10:26:24
import os, io, locale, math, random, datetime

from collections import OrderedDict
from microapp import App
from fortlab.kggenfile import (
    genkobj,
    gensobj,
    KERNEL_ID_0,
    event_register,
    create_rootnode,
    create_programnode,
    init_plugins,
    append_program_in_root,
    set_indent,
    plugin_config,
)
from fortlab.kgutils import (
    ProgramException,
    UserException,
    remove_multiblanklines,
    run_shcmd,
    tounicode,
)
from fortlab.resolver.kgparse import KGGenType
from fortlab.kgextra import (
    kgen_utils_file_head,
    kgen_utils_file_checksubr,
    kgen_get_newunit,
    kgen_error_stop,
    kgen_utils_file_tostr,
    kgen_utils_array_sumcheck,
    kgen_rankthread,
)

here = os.path.abspath(os.path.realpath(os.path.dirname(__file__)))
KGUTIL = "kgen_utils.f90"


class FortranKernelGenerator(App):
    _name_ = "kernelgen"
    _version_ = "0.1.0"

    def __init__(self, mgr):
        self.add_argument("analysis", help="analysis object")
        self.add_argument("--outdir", help="output directory")
        self.add_argument("--model", help="model object")

        self.add_argument("--repr-etime", dest="repr_etime", action='append', type=str, help="Specifying elapsedtime representativeness feature flags")
        self.add_argument("--repr-papi", dest="repr_papi", action='append', type=str, help="Specifying papi counter representativeness feature flags")
        self.add_argument("--repr-code", dest="repr_code", action='append', type=str, help="Specifying code coverage representativeness feature flags")

        self.register_forward("kerneldir", help="kernel generation code directory")
        self.register_forward("statedir", help="state generation code directory")

    def _process_repr_flags(self, opts):

        if opts.repr_etime:
            for line in opts.repr_etime:
                line = line["_"]
                for eopt in line.split(','):
                    split_eopt = eopt.split('=', 1)
                    if len(split_eopt)==1:
                        if split_eopt[0] == 'enable':
                            self.config['model']['types']['etime']['enabled'] = True
                        elif split_eopt[0] == 'disable':
                            self.config['model']['types']['etime']['enabled'] = False
                        else:
                            raise UserException('Unknown elapsed-time flag option: %s' % eopt)
                    elif len(split_eopt)==2:

                        self.config['model']['types']['etime']['enabled'] = True

                        if split_eopt[0] in [ 'minval', 'maxval' ]:
                            self.config['model']['types']['etime'][split_eopt[0]] = float(split_eopt[1])
                        elif split_eopt[0] in ('nbins', 'ndata'):
                            self.config['model']['types']['etime'][split_eopt[0]] = int(split_eopt[1])
                        elif split_eopt[0] in ('timer', ):
                            self.config['model']['types']['etime'][split_eopt[0]] = split_eopt[1]
                        else:
                            raise UserException('Unknown elapsed-time flag option: %s' % eopt)

        if opts.repr_papi:
            for line in opts.repr_papi:
                line = line["_"]
                for popt in line.split(','):
                    split_popt = popt.split('=', 1)
                    if len(split_popt)==1:
                        if split_popt[0] == 'enable':
                            self.config['model']['types']['papi']['enabled'] = True
                        elif split_popt[0] == 'disable':
                            self.config['model']['types']['papi']['enabled'] = False
                        else:
                            raise UserException('Unknown papi-counter flag option: %s' % popt)
                    elif len(split_popt)==2:

                        self.config['model']['types']['papi']['enabled'] = True

                        if split_popt[0] in [ 'minval', 'maxval', 'header', 'static', 'dynamic', 'event' ]:
                            self.config['model']['types']['papi'][split_popt[0]] = split_popt[1]
                        elif split_popt[0] in ('nbins', 'ndata'):
                            self.config['model']['types']['papi'][split_popt[0]] = int(split_popt[1])
                        else:
                            raise UserException('Unknown papi-counter flag option: %s' % popt)
        if opts.repr_code:
            for line in opts.repr_code:
                line = line["_"]
                for copt in line.split(','):
                    split_copt = copt.split('=', 1)
                    if len(split_copt)==1:
                        if split_copt[0] == 'enable':
                            self.config['model']['types']['code']['enabled'] = True
                        elif split_copt[0] == 'disable':
                            self.config['model']['types']['code']['enabled'] = False
                        else:
                            raise UserException('Unknown code-coverage flag option: %s' % copt)
                    elif len(split_copt)==2:

                        self.config['model']['types']['code']['enabled'] = True

                        if split_copt[0] in [ 'percentage' ]:
                            self.config['model']['types']['code'][split_copt[0]] = float(split_copt[1])
                        elif split_copt[0] in [ 'filter' ]:
                            self.config['model']['types']['code'][split_copt[0]] = split_copt[1].strip().split(':')
                        elif split_copt[0] in ( 'ndata' ):
                            self.config['model']['types']['code'][split_popt[0]] = int(split_copt[1])
                        else:
                            raise UserException('Unknown code-coverage flag option: %s' % copt)

            # enable coverage feature at extractor
            #if self.config['model']['types']['code']['enabled']:
            #    self.config['plugin']['priority']['ext.coverage'] = '%s/plugins/coverage'%KGEN_EXT

    def perform(self, args):

        self.config = args.analysis["_"]

        # process representativeness flags
        self._process_repr_flags(args)

        args.outdir = args.outdir["_"] if args.outdir else os.getcwd()

        if not os.path.exists(args.outdir):
            os.makedirs(args.outdir)

        self._trees = []
        self.genfiles = []

        self.config["used_srcfiles"].clear()

        state_realpath = os.path.realpath(os.path.join(args.outdir, "state"))
        kernel_realpath = os.path.realpath(os.path.join(args.outdir, "kernel"))

        self.config["path"]["kernel_output"] = kernel_realpath
        self.config["path"]["state_output"] = state_realpath

        self.add_forward(kerneldir=kernel_realpath, statedir=state_realpath)

        if not os.path.exists(kernel_realpath):
            os.makedirs(kernel_realpath)

        if not os.path.exists(state_realpath):
            os.makedirs(state_realpath)

        state_makefile = os.path.realpath(os.path.join(state_realpath, "Makefile"))
        kernel_makefile = os.path.realpath(os.path.join(kernel_realpath, "Makefile"))
        gencore_plugindir = os.path.join(here, "plugins", "gencore")
        verify_plugindir = os.path.join(here, "plugins", "verification")
        timing_plugindir = os.path.join(here, "plugins", "simple_timing")
        perturb_plugindir = os.path.join(here, "plugins", "perturb")

        if args.model:
            try:
                if os.path.exists(args.model["_"]):
                    with open(args.model["_"]) as f:
                        self.read_model(json.load(f), self.config)
            except:
                self.read_model(args.model["_"], self.config)
        else:
            self.config["invocation"]["triples"].append(
                (("0", "0"), ("0", "0"), ("1", "1"))
            )

        plugins = (
            ("ext.gencore", gencore_plugindir),
            ("ext.verification", verify_plugindir),
            ("ext.simple_timing", timing_plugindir),
            ("ext.perturb", perturb_plugindir),
        )

        init_plugins([KERNEL_ID_0], plugins, self.config)
        plugin_config["current"].update(self.config)

        driver = create_rootnode(KERNEL_ID_0)
        self._trees.append(driver)
        program = create_programnode(driver, KERNEL_ID_0)
        program.name = self.config["kernel_driver"]["name"]
        append_program_in_root(driver, program)

        for filepath, (srcobj, mods_used, units_used) in self.config[
            "srcfiles"].items():
            if hasattr(srcobj.tree, "geninfo") and KGGenType.has_state(
                srcobj.tree.geninfo
            ):
                kfile = genkobj(None, srcobj.tree, KERNEL_ID_0)
                sfile = gensobj(None, srcobj.tree, KERNEL_ID_0)
                sfile.kgen_stmt.used4genstate = False
                if kfile is None or sfile is None:
                    raise kgutils.ProgramException(
                        "Kernel source file is not generated for %s." % filepath
                    )
                self.genfiles.append((kfile, sfile, filepath))
                self.config["used_srcfiles"][filepath] = (
                    kfile,
                    sfile,
                    mods_used,
                    units_used,
                )

        for plugin_name in event_register.keys():
            if not plugin_name.startswith("ext"):
                continue
            for kfile, sfile, filepath in self.genfiles:
                kfile.created([plugin_name])
                sfile.created([plugin_name])

            for tree in self._trees:
                tree.created([plugin_name])

        for plugin_name in event_register.keys():
            if not plugin_name.startswith("ext"):
                continue
            for kfile, sfile, filepath in self.genfiles:
                kfile.process([plugin_name])
                sfile.process([plugin_name])

            for tree in self._trees:
                tree.process([plugin_name])

        for plugin_name in event_register.keys():
            if not plugin_name.startswith("ext"):
                continue
            for kfile, sfile, filepath in self.genfiles:
                kfile.finalize([plugin_name])
                sfile.finalize([plugin_name])

            for tree in self._trees:
                tree.finalize([plugin_name])

        for plugin_name in event_register.keys():
            if not plugin_name.startswith("ext"):
                continue
            for kfile, sfile, filepath in self.genfiles:
                kfile.flatten(KERNEL_ID_0, [plugin_name])
                sfile.flatten(KERNEL_ID_0, [plugin_name])

            for tree in self._trees:
                tree.flatten(KERNEL_ID_0, [plugin_name])

        kernel_files = []
        state_files = []
        enc = locale.getpreferredencoding(False)

        for kfile, sfile, filepath in self.genfiles:
            filename = os.path.basename(filepath)
            set_indent("")
            klines = kfile.tostring()
            if klines is not None:
                klines = remove_multiblanklines(klines)
                kernel_files.append(filename)
                with io.open(
                    os.path.join(kernel_realpath, filename), "w", encoding=enc
                ) as (fd):
                    fd.write(tounicode(klines))
            if sfile.kgen_stmt.used4genstate:
                set_indent("")
                slines = sfile.tostring()
                if slines is not None:
                    slines = remove_multiblanklines(slines)
                    state_files.append(filename)
                    with io.open(
                        os.path.join(state_realpath, filename), "w", encoding=enc
                    ) as (fd):
                        fd.write(tounicode(slines))

        with io.open(
            os.path.join(
                kernel_realpath, "%s.f90" % self.config["kernel_driver"]["name"]
            ),
            "w",
            encoding=enc,
        ) as (fd):
            set_indent("")
            lines = driver.tostring()
            if lines is not None:
                lines = remove_multiblanklines(lines)
                fd.write(tounicode(lines))

        kernel_files.append(self.config["kernel"]["name"])
        kernel_files.append(KGUTIL)
        self.generate_kgen_utils(kernel_realpath, enc)
        self.generate_kernel_makefile(kernel_realpath, enc)
        kernel_files.append("Makefile")
        self.generate_state_makefile(state_realpath, enc)
        state_files.append("Makefile")

        if self.config["cmd_clean"]["cmds"]:
            run_shcmd(self.config["cmd_clean"]["cmds"])

        if self.config["state_switch"]["clean"]:
            run_shcmd(self.config["state_switch"]["clean"])

        out, err, retcode = run_shcmd("make", cwd=state_realpath)

        if retcode != 0:
            print("ERROR: state make error")

        out, err, retcode = run_shcmd("make recover", cwd=state_realpath)

        if retcode != 0:
            print("ERROR: state recover error")

        if self.config["state_switch"]["clean"]:
            run_shcmd(self.config["state_switch"]["clean"])

        return

    def generate_kgen_utils(self, kernel_path, enc):
        with io.open(os.path.join(kernel_path, KGUTIL), "w", encoding=enc) as (f):
            f.write(tounicode("MODULE kgen_utils_mod"))
            f.write(tounicode(kgen_utils_file_head))
            f.write(tounicode("\n"))
            f.write(tounicode("CONTAINS"))
            f.write(tounicode("\n"))
            f.write(tounicode(kgen_utils_array_sumcheck))
            f.write(tounicode(kgen_utils_file_tostr))
            f.write(tounicode(kgen_utils_file_checksubr))
            f.write(tounicode(kgen_get_newunit))
            f.write(tounicode(kgen_error_stop))
            f.write(tounicode(kgen_rankthread))
            f.write(tounicode("END MODULE kgen_utils_mod\n"))

    def write(self, f, line, n=True, t=False):
        nl = ""
        tab = ""
        if n:
            nl = "\n"
        if t:
            tab = "\t"
        f.write(tounicode(tab + line + nl))

    def obj(self, file):
        l = file.split(".")
        if len(l) > 1:
            l[-1] = "o"
        return (".").join(l)

    def generate_kernel_makefile(self, kernelpath, enc):

        openmp_flags = ["-fopenmp", "-qopenmp", "-mp", "-xopenmp", "-qsmp=omp"]
        callsite_base = os.path.basename(self.config["callsite"]["filepath"])
        driver_base = "%s.f90" % self.config["kernel_driver"]["name"]

        dep_base_srcfiles = [
            os.path.basename(filepath)
            for filepath, srclist in self.config["used_srcfiles"].items()
        ]
        dep_bases = dep_base_srcfiles + [driver_base]

        all_objs_srcfiles = [
            self.obj(dep_base_srcfile) for dep_base_srcfile in dep_base_srcfiles
        ]
        all_objs = all_objs_srcfiles + [self.obj(driver_base), self.obj(KGUTIL)]

        depends = OrderedDict()
        depends[driver_base] = (" ").join(all_objs_srcfiles + [self.obj(KGUTIL)])

        for filepath, (kfile, sfile, mods_used, units_used) in self.config[
            "used_srcfiles"
        ].items():
            dep = [self.obj(KGUTIL)]
            for mod in mods_used:
                if (
                    mod.reader.id != filepath
                    and self.obj(os.path.basename(mod.reader.id)) not in dep
                ):
                    dep.append(self.obj(os.path.basename(mod.reader.id)))

            for unit in units_used:
                if (
                    unit.item.reader.id != filepath
                    and self.obj(os.path.basename(unit.item.reader.id)) not in dep
                ):
                    dep.append(self.obj(os.path.basename(unit.item.reader.id)))

            basename = os.path.basename(filepath)
            if basename == callsite_base:
                dobjs = all_objs[:]
                dobjs.remove(self.obj(callsite_base))
                dobjs.remove(self.obj(driver_base))
                depends[basename] = (" ").join(dobjs)
            else:
                depends[basename] = (" ").join(dep)

        compilers = OrderedDict()
        compiler_options = OrderedDict()

        for path, kfile in self.config["include"]["file"].items():
            base = os.path.basename(path)

            if base not in dep_bases:
                continue

            comp = None

            if "compiler" in kfile:
                comp = kfile["compiler"]
            elif "compiler" in self.config["include"]["compiler"]:
                comp = self.config["include"]["compiler"]["compiler"]

            if comp:
                if comp in compilers:
                    if base not in compilers[comp]:
                        compilers[comp].append(base)
                else:
                    compilers[comp] = [base, driver_base]

            opts = ""

            if "compiler_options" in self.config["include"]["compiler"]:
                opts = (
                    opts + " " + self.config["include"]["compiler"]["compiler_options"]
                )

            #if "compiler_options" in kfile and kfile["compiler_options"]:
            #    opts = opts + " " + kfile["compiler_options"]

            if "options" in kfile and kfile["options"]:
                opts = opts + " " + " ".join(kfile["options"])

            if self.config["model"]["types"]["code"]["enabled"]:
                if comp:
                    compname = os.path.basename(comp)
                    if compname == "ifort":
                        opts += " -fpp "
                    elif compname == "gfortran":
                        opts += " -cpp "
                    elif compname.startswith("pg"):
                        opts += " -Mpreprocess "

                opts += ' -D KGEN_COVERAGE # Comment out "-D KGEN_COVERAGE" to turn off coverage feature.'

            if self.config["add_mpi_frame"]["enabled"]:
                if comp:
                    compname = os.path.basename(comp)

                    if compname == "ifort" and "-fpp" not in opts:
                        opts += " -fpp "
                    elif compname == "gfortran" and "-cpp" not in opts:
                        opts += " -cpp "
                    elif compname.startswith("pg") and "-Mpreprocess" not in opts:
                        opts += " -Mpreprocess "

                    opts += " -D_MPI "

            if len(opts) > 0:
                if opts in compiler_options:
                    if base not in compiler_options[opts]:
                        compiler_options[opts].append(base)
                else:
                    compiler_options[opts] = [base, driver_base]

        link_flags = (" ").join(self.config["kernel_option"]["linker"]["add"])
        objects = ""

        if "import" in self.config["include"]:
            for path, import_type in self.config["include"]["import"].items():
                if import_type == "library" or import_type == "shared_library":
                    inc = "-L" + path
                    pos1 = import_type.find("(")
                    pos2 = import_type.find(")")
                    lib = "-l" + import_type[pos1 + 1 : pos2].strip()
                    link_flags += " %s %s" % (inc, lib)

                elif import_type == "static-library":
                    link_flags += " %s" % path

                elif import_type == "object":
                    objects += " %s" % os.path.basename(path)
                    shutil.copy(
                        path,
                        "%s/%s"
                        % (
                            self.config["path"]["outdir"],
                            self.config["path"]["kernel"],
                        ),
                    )

        with io.open(os.path.join(kernelpath, "Makefile"), "w", encoding=enc) as (f):
            self.write(f, "# Makefile for KGEN-generated kernel")
            self.write(f, "")

            if self.config["kernel_option"]["FC"]:
                self.write(f, "FC_0 := %s" % self.config["kernel_option"]["FC"])

            else:
                if self.config["add_mpi_frame"]["enabled"]:
                    self.write(f, "# Originally used compiler(s)")
                    for i, compiler in enumerate(compilers):
                        self.write(f, "#FC_%d := %s" % (i, compiler))

                    self.write(f, "FC_0 := %s" % self.config["add_mpi_frame"]["mpifc"])
                else:
                    for i, compiler in enumerate(compilers):
                        self.write(f, "FC_%d := %s" % (i, compiler))

                if self.config["kernel_option"]["FC_FLAGS"]:
                    self.write( f, "FC_FLAGS_SET_0 := %s" %
                                self.config["kernel_option"]["FC_FLAGS"],)
                else:
                    for i, options in enumerate(compiler_options):
                        opt_list = options.split()
                        L = len(opt_list)
                        new_options = []

                        if L > 1:
                            skip_next = False
                            for opt in opt_list:
                                if skip_next:
                                    skip_next = False
                                    continue
                                if (opt in self.config["kernel_option"]["compiler"]["remove"]):
                                    pass

                                elif ("%s+" % opt in
                                        self.config["kernel_option"]["compiler"][ "remove" ]):
                                    skip_next = True

                                elif opt not in openmp_flags:
                                    new_options.append(opt)

                        for add_opt in self.config["kernel_option"]["compiler"]["add"]:
                            if add_opt not in new_options:
                                new_options.append(add_opt)

                        self.write(f, "FC_FLAGS_SET_%d := %s" %
                                    (i, (" ").join(new_options)))

            prerun_build_str = ""
            if (
                self.config["prerun"]["kernel_build"]
                or self.config["prerun"]["build"]
                or self.config["prerun"]["kernel_run"]
                or self.config["prerun"]["run"]
            ):
                self.write(
                    f,
                    "# Comment-out below if your system environment is different from one that this kernel is extracted from.",
                )
                self.write(
                    f,
                    "# NOTE: the environment is valid only if you are using this kernel on the same computing system that the kernel was extracted from.",
                )
            if self.config["prerun"]["kernel_build"]:
                self.write(
                    f, "PRERUN_BUILD := %s; " % self.config["prerun"]["kernel_build"]
                )
                prerun_build_str = "${PRERUN_BUILD}"
            elif self.config["prerun"]["build"]:
                self.write(f, "PRERUN_BUILD := %s; " % self.config["prerun"]["build"])
                prerun_build_str = "${PRERUN_BUILD}"
            prerun_run_str = ""
            if self.config["prerun"]["kernel_run"]:
                self.write(
                    f, "PRERUN_RUN := %s; " % self.config["prerun"]["kernel_run"]
                )
                prerun_run_str = "${PRERUN_RUN}"
            else:
                if self.config["prerun"]["run"]:
                    self.write(f, "PRERUN_RUN := %s; " % self.config["prerun"]["run"])
                    prerun_run_str = "${PRERUN_RUN}"
                self.write(f, "")
                self.write(f, "ALL_OBJS := %s" % (" ").join(all_objs))
                self.write(f, "")
                if self.config["model"]["types"]["papi"]["enabled"]:
                    self.write(
                        f,
                        "PAPI_EVENT := %s"
                        % self.config["model"]["types"]["papi"]["event"],
                    )
                    self.write(f, "")
                self.write(f, "build: ${ALL_OBJS}")
                fc_str = "FC_0"
                fc_flags_str = "FC_FLAGS_SET_0"
                self.write(
                    f,
                    "%s${%s} ${%s} -o kernel.exe $^ %s %s"
                    % (prerun_build_str, fc_str, fc_flags_str, link_flags, objects),
                    t=True,
                )
                self.write(f, "")
                self.write(f, "run: build")
                if self.config["add_mpi_frame"]["enabled"]:
                    self.write(
                        f,
                        "%s%s -np %s ./kernel.exe"
                        % (
                            prerun_run_str,
                            self.config["add_mpi_frame"]["mpirun"],
                            self.config["add_mpi_frame"]["np"],
                        ),
                        t=True,
                    )
                else:
                    self.write(f, "%s./kernel.exe" % prerun_run_str, t=True)
                self.write(f, "")
                if self.config["model"]["types"]["papi"]["enabled"]:
                    self.write(f, "papi: build-papi")
                    if self.config["add_mpi_frame"]["enabled"]:
                        self.write(
                            f,
                            "%s%s -np %s ./kernel.exe"
                            % (
                                prerun_run_str,
                                self.config["add_mpi_frame"]["mpirun"],
                                self.config["add_mpi_frame"]["np"],
                            ),
                            t=True,
                        )
                    else:
                        self.write(f, "%s./kernel.exe" % prerun_run_str, t=True)
                    self.write(f, "")
                if self.config["model"]["types"]["papi"]["enabled"]:
                    self.write(f, "build-papi: ${ALL_OBJS}")
                    fc_str = "FC_0"
                    fc_flags_str = "FC_FLAGS_SET_0"
                    if self.config["model"]["types"]["papi"]["static"] is not None:
                        link_flags += (
                            " %s" % self.config["model"]["types"]["papi"]["static"]
                        )
                    if self.config["model"]["types"]["papi"]["dynamic"] is not None:
                        ddir, dlib = os.path.split(
                            self.config["model"]["types"]["papi"]["dynamic"]
                        )
                        root, ext = dlib.split(".", 1)
                        if len(root) > 3:
                            link_flags += " -L%s -l%s" % (ddir, root[3:])
                    self.write(
                        f,
                        "%s${%s} ${%s} -o kernel.exe $^ %s %s"
                        % (prerun_build_str, fc_str, fc_flags_str, link_flags, objects),
                        t=True,
                    )
                    self.write(f, "")
                for dep_base in dep_bases:
                    self.write(
                        f,
                        "%s: %s %s" % (self.obj(dep_base), dep_base, depends[dep_base]),
                    )
                    dfc_str = "FC"
                    dfc_flags_str = "FC_FLAGS"
                    if self.config["kernel_option"]["FC"] is None:
                        for i, (compiler, files) in enumerate(compilers.items()):
                            if dep_base in files:
                                dfc_str += "_%d" % i
                                break

                    if self.config["kernel_option"]["FC_FLAGS"] is None:
                        for i, (compiler_option, files) in enumerate(
                            compiler_options.items()
                        ):
                            if dep_base in files:
                                dfc_flags_str += "_SET_%d" % i
                                break

                    if dfc_str == "FC" or self.config["add_mpi_frame"]["enabled"]:
                        dfc_str = "FC_0"
                    if dfc_flags_str == "FC_FLAGS":
                        dfc_flags_str = "FC_FLAGS_SET_0"
                    if (
                        self.config["model"]["types"]["papi"]["enabled"]
                        and self.config["model"]["types"]["papi"]["header"] is not None
                        and dep_base in [callsite_base, driver_base]
                    ):
                        self.write(f, "ifeq (${MAKECMDGOALS}, papi)")
                        papi_flags_str = (
                            " -DKGEN_PAPI -I%s"
                            % os.path.split(
                                self.config["model"]["types"]["papi"]["header"]
                            )[0]
                        )
                        self.write(
                            f,
                            '%s %s %s $< | sed "s/KGENPAPIEVENT/${PAPI_EVENT}/g" > tmp.$<'
                            % (
                                self.config["bin"]["pp"],
                                self.config["bin"]["cpp_flags"],
                                papi_flags_str,
                            ),
                            t=True,
                        )
                        self.write(f, "else")
                        self.write(
                            f,
                            "%s %s $< tmp.$<"
                            % (
                                self.config["bin"]["pp"],
                                self.config["bin"]["cpp_flags"],
                            ),
                            t=True,
                        )
                        self.write(f, "endif")
                        self.write(
                            f,
                            "%s${%s} ${%s} -c -o $@ tmp.$<"
                            % (prerun_build_str, dfc_str, dfc_flags_str),
                            t=True,
                        )
                        self.write(f, "rm -f tmp.$<", t=True)
                    else:
                        self.write(
                            f,
                            "%s${%s} ${%s} -c -o $@ $<"
                            % (prerun_build_str, dfc_str, dfc_flags_str),
                            t=True,
                        )
                    self.write(f, "")

            self.write(f, "%s: %s" % (self.obj(KGUTIL), KGUTIL))
            self.write(
                f,
                "%s${%s} ${%s} -c -o $@ $<" % (prerun_build_str, fc_str, fc_flags_str),
                t=True,
            )
            self.write(f, "")
            self.write(
                f,
                "%s${%s} ${%s} -c -o $@ $<" % (prerun_build_str, fc_str, fc_flags_str),
                t=True,
            )
            self.write(f, "")
            self.write(f, "clean:")
            self.write(f, "rm -f kernel.exe *.mod ${ALL_OBJS}", t=True)
        return

    def generate_state_makefile(self, statepath, enc):

        org_files = [
            filepath
            for filepath, (kfile, sfile, mods_used, units_used) in self.config[
                "used_srcfiles"
            ].items()
            if sfile.kgen_stmt.used4genstate
        ]
        if self.config["topblock"]["stmt"].reader.id not in org_files:
            org_files.append(self.config["topblock"]["filepath"])
        with io.open(os.path.join(statepath, "Makefile"), "w", encoding=enc) as (f):
            self.write(f, "# Makefile for KGEN-generated instrumentation")
            self.write(f, "")
            cwd = os.path.realpath(self.config["cwd"])
            prerun_build_str = ""
            if self.config["prerun"]["build"]:
                self.write(f, "PRERUN_BUILD := %s" % self.config["prerun"]["build"])
                prerun_build_str = "${PRERUN_BUILD}; "
            prerun_run_str = ""
            if self.config["prerun"]["run"]:
                self.write(f, "PRERUN_RUN := %s" % self.config["prerun"]["run"])
                prerun_run_str = "${PRERUN_RUN}; "
            self.write(f, "")
            if len(self.config["cmd_run"]["cmds"]) > 0:
                self.write(f, "run: build")
                self.write(
                    f,
                    "%scd %s; %s"
                    % (prerun_run_str, cwd, self.config["cmd_run"]["cmds"]),
                    t=True,
                )
            else:
                self.write(
                    f,
                    "echo \"No information is provided to run. Please specify run commands using 'state-run' command line option\"; exit -1",
                    t=True,
                )
            self.write(f, "")
            if len(self.config["cmd_build"]["cmds"]) > 0:
                self.write(f, "build: %s" % self.config["state_switch"]["type"])
                self.write(
                    f,
                    "%scd %s; %s"
                    % (prerun_build_str, cwd, self.config["cmd_build"]["cmds"]),
                    t=True,
                )
                for org_file in org_files:
                    self.write(
                        f, "mv -f %(f)s.kgen_org %(f)s" % {"f": org_file}, t=True
                    )

            else:
                self.write(
                    f,
                    "echo \"No information is provided to build. Please specify build commands using 'state-build' command line option\"; exit -1",
                    t=True,
                )
            self.write(f, "")
            self.write(f, "%s: save" % self.config["state_switch"]["type"])
            if self.config["state_switch"]["type"] == "replace":
                for org_file in org_files:
                    basename = os.path.basename(org_file)
                    self.write(f, "rm -f %s" % org_file, t=True)
                    self.write(
                        f,
                        "cp -f %(f1)s %(f2)s" % {"f1": basename, "f2": org_file},
                        t=True,
                    )

            elif self.config["state_switch"]["type"] == "copy":
                for org_file in org_files:
                    basename = os.path.basename(org_file)
                    self.write(
                        f,
                        "rm -f %s/%s"
                        % (self.config["state_switch"]["directory"], basename),
                        t=True,
                    )
                    self.write(
                        f,
                        "cp -f %s/%s/%s %s"
                        % (
                            self.config["path"]["outdir"],
                            self.config["path"]["state"],
                            basename,
                            self.config["state_switch"]["directory"],
                        ),
                        t=True,
                    )

            self.write(f, "")
            self.write(f, "recover:")
            for org_file in org_files:
                self.write(f, "rm -f %s" % org_file, t=True)
                self.write(
                    f,
                    "cp -f %s.kgen_org %s" % (os.path.basename(org_file), org_file),
                    t=True,
                )

            self.write(f, "")
            self.write(f, "recover_from_srcdir:")
            for org_file in org_files:
                self.write(f, "rm -f %s" % org_file, t=True)
                self.write(f, "cp -f %(f)s.kgen_org %(f)s" % {"f": org_file}, t=True)

            self.write(f, "")
            self.write(f, "save:")
            for org_file in org_files:
                self.write(
                    f,
                    "if [ ! -f %(f)s.kgen_org ]; then cp -f %(f)s %(f)s.kgen_org; fi"
                    % {"f": org_file},
                    t=True,
                )
                self.write(
                    f,
                    "if [ ! -f %(g)s.kgen_org ]; then cp -f %(f)s %(g)s.kgen_org; fi"
                    % {"f": org_file, "g": os.path.basename(org_file)},
                    t=True,
                )

            self.write(f, "")
            if len(self.config["cmd_clean"]["cmds"]) > 0:
                self.write(f, "clean:")
                self.write(f, "%s" % self.config["cmd_clean"]["cmds"], t=True)
            self.write(f, "")

    def read_model(self, modeljson, config):

        if "etime" not in modeljson:
            raise UserException("'etime' is not in model json file")

        emeta = modeljson["etime"]["_summary_"]
        del modeljson["etime"]["_summary_"]
        etimes = modeljson["etime"]

        try:
            etimemin = emeta["elapsedtime_min"]
            etimemax = emeta["elapsedtime_max"]
            netimes = emeta["number_eitmes"]
            etimediff = etimemax - etimemin
            etimeres = emeta["elapsedtime_res"]
            if etimediff == 0:
                nbins = 1
            else:
                nbins = max(min(config["model"]["types"]["etime"]["nbins"], netimes), 2)
            if nbins > 1:
                etimebins = [{} for _ in range(nbins)]
                etimecounts = [0 for _ in range(nbins)]
            else:
                etimebins = [{}]
                etimecounts = [0]
            idx = 0
            for ranknum, d1 in etimes.items():
                for threadnum, d2 in d1.items():
                    for invokenum, (start, stop) in d2.items():

            #for opt in cfg.options("elapsedtime.elapsedtime"):
            #    ranknum, threadnum, invokenum = tuple(num for num in opt.split())
            #    start, stop = cfg.get("elapsedtime.elapsedtime", opt).split(",")
                        estart = float(start)
                        eend = float(stop)
                        etimeval = eend - estart
                        if nbins > 1:
                            binnum = int(
                                math.floor((etimeval - etimemin) / etimediff * (nbins - 1))
                            )
                        else:
                            binnum = 0
                        etimecounts[binnum] += 1
                        invokenum = int(invokenum)
                        if invokenum not in etimebins[binnum]:
                            etimebins[binnum][invokenum] = {}
                        if ranknum not in etimebins[binnum][invokenum]:
                            etimebins[binnum][invokenum][ranknum] = {}
                        if threadnum not in etimebins[binnum][invokenum][ranknum]:
                            etimebins[binnum][invokenum][ranknum][threadnum] = float(etimeval)
                        else:
                            raise Exception(
                                "Dupulicated data: (%s, %s, %s, %s)"
                                % (invokenum, ranknum, threadnum, etimeval)
                            )
                        idx += 1
                        if idx % 100000 == 0:
                            print(
                                "Processed %d items: %s"
                                % (
                                    idx,
                                    datetime.datetime.now().strftime("%I:%M%p on %B %d, %Y"),
                                )
                            )

        except Exception as e:
            raise Exception("Please check the format of elapsedtime file: %s" % str(e))

        totalcount = sum(etimecounts)
        countdist = [float(count) / float(totalcount) for count in etimecounts]
        ndata = config["model"]["types"]["etime"]["ndata"]
        datacollect = [int(round(dist * ndata)) for dist in countdist]
        triples = []
        for binnum, etimebin in enumerate(etimebins):
            bin_triples = []
            range_begin = (
                binnum * (etimemax - etimemin) / nbins + etimemin
                if binnum > 0
                else etimemin
            )
            range_end = (
                (binnum + 1) * (etimemax - etimemin) / nbins + etimemin
                if binnum < nbins - 1
                else None
            )
            bunit = "sec"
            if range_begin < 1e-06:
                bunit = "usec"
                range_begin *= 1000000.0
            if range_end is None:
                print(
                    "From bin # %d [ %f (%s) ~ ] %f %% of %d"
                    % (binnum, range_begin, bunit, countdist[binnum] * 100, totalcount)
                )
            else:
                eunit = "sec"
                if range_end < 1e-06:
                    eunit = "usec"
                    range_end *= 1000000.0
                print(
                    "From bin # %d [ %f (%s) ~ %f (%s) ] %f %% of %d"
                    % (
                        binnum,
                        range_begin,
                        bunit,
                        range_end,
                        eunit,
                        countdist[binnum] * 100,
                        totalcount,
                    )
                )
            for invokenum in sorted(etimebin.keys()):
                if len(bin_triples) >= datacollect[binnum]:
                    break
                bininvokes = list(etimebin[invokenum].keys())
                random.shuffle(bininvokes)
                for ranknum in bininvokes:
                    if len(bin_triples) >= datacollect[binnum]:
                        break
                    binranks = list(etimebin[invokenum][ranknum].keys())
                    random.shuffle(binranks)
                    for threadnum in binranks:
                        bin_triples.append((ranknum, threadnum, invokenum))
                        print(
                            "        invocation triple: %s:%s:%s"
                            % (ranknum, threadnum, invokenum)
                        )

            triples.extend(bin_triples)

        print("Number of bins: %d" % nbins)
        print("Minimun elapsed time: %f" % etimemin)
        print("Maximum elapsed time: %f" % etimemax)
        for ranknum, threadnum, invokenum in triples:
            config["invocation"]["triples"].append(
                (
                    (str(ranknum), str(ranknum)),
                    (str(threadnum), str(threadnum)),
                    (str(invokenum), str(invokenum)),
                )
            )

        return


# okay decompiling kernelgen.pyc



# kgen_compiler.py

import os
import re

class GenericCompiler(object):
    compid = None
    compnames = []
    openmp = None
    file_exts = None
    fpp = ''

    discard_opts_noarg = [ ]
    discard_opts_arg = [ ]

    def get_discard_opts_noarg(self):
        return [ '-c' ]

    def get_discard_opts_arg(self):
        return [ '-o', '-l', '-L', '-W' ]

    def _getmacro(self, macro):
        splitmacro = macro.split('=')
        if len(splitmacro)==1:
           return (splitmacro[0], None)
        elif len(splitmacro)==2:
            return tuple(splitmacro)
        else:
            raise

    def parse_option(self, options, pwd):
        incs = []
        macros = []
        openmp = []
        srcs = []
        flags = []

        discard_flag = False

        iflag = False
        dflag = False

        for item in options[1:]:

            if discard_flag:
                discard_flag = False
                continue

            if iflag:
                for p in item.split(':'):
                    if p[0]=='/':
                        incs.append(p)
                    else:
                        incs.append(os.path.realpath('%s/%s'%(pwd,p)))
                iflag = False
                continue
            if dflag:
                macros.append(self._getmacro(item))
                dflag = False
                continue

            if item.startswith('-I'):
                if len(item)>2:
                    for p in item[2:].split(':'):
                        if p[0]=='/':
                            incs.append(p)
                        else:
                            incs.append(os.path.realpath('%s/%s'%(pwd,p)))
                else: iflag = True
            elif item.startswith('-D'):
                if len(item)>2:
                    macros.append(self._getmacro(item[2:]))
                else: dflag = True
            elif item.startswith('-WF,-D'):
                if len(item)>6:
                    macros.append(self._getmacro(item[6:]))
                else: dflag = True
            elif self.openmp_opt and any(not re.match(r'%s\Z'%pattern, item) is None for pattern in self.openmp_opt):
                    openmp.append(item)
            elif item.startswith('-'):
                if item in self.get_discard_opts_noarg():
                    pass
                elif any( item.startswith(f) for f in self.get_discard_opts_arg() ):
                    for f in self.get_discard_opts_arg():
                        if item==f:
                            if len(item)==len(f):
                                discard_flag = True
                            break
                else:
                    flags.append(item)
            elif self.file_exts and '.' in item and item.split('.')[-1] in self.file_exts:
                if item[0]=='/':
                    srcs.append(item)
                else:
                    srcs.append(os.path.realpath('%s/%s'%(pwd,item)))
            else:
                flags.append(item)

        if len(srcs)>0:
            return (srcs, incs, macros, openmp, flags)
        else:
            return ([], [], [], [], [])

class GenericFortranCompiler(GenericCompiler):
    file_exts = ['f', 'f90', 'f95', 'f03', 'f08', 'ftn', 'F', 'F90', 'F95', 'F03', 'F08', 'FTN' ]

    def get_discard_opts_noarg(self):
        return super(GenericFortranCompiler, self).get_discard_opts_noarg()

    def get_discard_opts_arg(self):
        return super(GenericFortranCompiler, self).get_discard_opts_arg()

class IntelFortranCompiler(GenericFortranCompiler):
    compid = 'ifort'
    compnames = ['ifort']
    openmp_opt = [ r'-openmp' ]
    fpp = '-fpp'

    discard_opts_arg = [ '-module' ]

    def get_discard_opts_noarg(self):
        return super(IntelFortranCompiler, self).get_discard_opts_noarg() + self.discard_opts_noarg

    def get_discard_opts_arg(self):
        return super(IntelFortranCompiler, self).get_discard_opts_arg() + self.discard_opts_arg

class GnuFortranCompiler(GenericFortranCompiler):
    compid = 'gfortran'
    compnames = ['gfortran']
    openmp_opt = [ r'-fopenmp' ]
    fpp = '-cpp'

    discard_opts_arg = [ '-J' ]

    def get_discard_opts_noarg(self):
        return super(GnuFortranCompiler, self).get_discard_opts_noarg() + self.discard_opts_noarg

    def get_discard_opts_arg(self):
        return super(GnuFortranCompiler, self).get_discard_opts_arg() + self.discard_opts_arg

class PgiFortranCompiler(GenericFortranCompiler):
    compid = 'pgfortran'
    compnames = ['pgfortran', 'pgf77', 'pgf90', 'pgf95', 'pghpf']
    openmp_opt = [ r'-mp' ]
    fpp = '-Mpreprocess'

    discard_opts_arg = [ '-module' ]

    def get_discard_opts_noarg(self):
        return super(PgiFortranCompiler, self).get_discard_opts_noarg() + self.discard_opts_noarg

    def get_discard_opts_arg(self):
        return super(PgiFortranCompiler, self).get_discard_opts_arg() + self.discard_opts_arg

class PathscaleFortranCompiler(GenericFortranCompiler):
    compid = 'path90'
    compnames = ['path90', 'path95']
    openmp_opt = [ r'-mp' ]

    def get_discard_opts_noarg(self):
        return super(PathscaleFortranCompiler, self).get_discard_opts_noarg() + self.discard_opts_noarg

    def get_discard_opts_arg(self):
        return super(PathscaleFortranCompiler, self).get_discard_opts_arg() + self.discard_opts_arg

class NagFortranCompiler(GenericFortranCompiler):
    compid = 'nagfor'
    compnames = ['nagfor']
    openmp_opt = [ r'-openmp' ]
    fpp = '-cpp'

    def get_discard_opts_noarg(self):
        return super(NagFortranCompiler, self).get_discard_opts_noarg() + self.discard_opts_noarg

    def get_discard_opts_arg(self):
        return super(NagFortranCompiler, self).get_discard_opts_arg() + self.discard_opts_arg

class IbmxlFortranCompiler(GenericFortranCompiler):

    compid = 'xlf'
    compnames = ['xlf', 'xlf90', 'xlf95', 'xlf2003', 'xlf_r', 'xlf90_r',
                 'xlf95_r', 'xlf2003_r', 'xlf2008', 'xlf2008_r', 'fort77']
    openmp_opt = [ r'-qsmp' ]
    fpp = '-WF,-qfpp'

    def get_discard_opts_noarg(self):
        return super(IbmxlFortranCompiler, self).get_discard_opts_noarg() + self.discard_opts_noarg

    def get_discard_opts_arg(self):
        return super(IbmxlFortranCompiler, self).get_discard_opts_arg() + self.discard_opts_arg

class CrayFortranCompiler(GenericFortranCompiler):
    compid = 'crayftn'
    compnames = ['crayftn', 'ftn']
    openmp_opt = [ r'-omp', r'-h\s+omp' ]
    fpp = '-e F'

    discard_opts_arg = [ '-J' ]

    def get_discard_opts_noarg(self):
        return super(CrayFortranCompiler, self).get_discard_opts_noarg() + self.discard_opts_noarg

    def get_discard_opts_arg(self):
        return super(CrayFortranCompiler, self).get_discard_opts_arg() + self.discard_opts_arg

class CompilerFactory(object):

    factories = {}

    @staticmethod
    def get_subclasses(cls):
        subclasses = set()
        work = [cls]
        while work:
            parent = work.pop()
            for child in parent.__subclasses__():
                if child not in subclasses:
                    subclasses.add(child)
                    work.append(child)
        return subclasses

    @staticmethod
    def createCompiler(compid):
        for subc in CompilerFactory.get_subclasses(GenericCompiler):
            if subc.compnames and compid in subc.compnames:
                if subc.compid not in CompilerFactory.factories:
                    CompilerFactory.factories[subc.compid] = subc()
                return CompilerFactory.factories[subc.compid]

#class IntelFortranCompiler(GenericFortranCompiler):
#    # space: False-no space, True-space required, None - any
#    # args: None-none, [|]-optional string, 
#    OPTIONS = { \
#        '-mmic': { 'args': None, 'default': False },
#        '-falias': { 'args': None, 'default': True },
#        '-fast': { 'args': None, 'default': False },
#        '-arch': { 'args': '[CORE-AVX2|CORE-AVX-I|AVX|SSE4.2|SSE4.1|SSSE3|SSE3|SSE2|SSE|IA32]', 'default': 'SSE2', 'space': True },
#        '-ax': { 'args': '[COMMON-AVX512|MIC-AVX512|CORE-AVX512|CORE-AVX2|CORE-AVX-I|AVX|SSE4.2|SSE4.1|SSSE3|SSE3|SSE2]', 'default': False, 'space': False },
#        '-m32': { 'args': None, 'default': False },
#        '-m64': { 'args': None, 'default': False },
#        '-xHost': { 'args': None, 'default': False },
#        '-ip': { 'args': None, 'default': False },
#        '-ipo': { 'args': '\d+', 'default': False, 'space': False },
#        '-no-ipo': { 'args': None, 'default': True },
#        '-mkl': { 'args': '[|=parallel|=sequential|=cluster]', 'default': False, 'space': False },
#        '-pad': { 'args': None, 'default': False },
#        '-no-pad': { 'args': None, 'default': False },
#        '-qopt-prefetch': { 'args': '[|=0|=1|=2|=3|=4]', 'default': False, 'space': False },
#        '-qno-opt-prefetch': { 'args': None, 'default': False },
#        '-simd': { 'args': None, 'default': True },
#        '-no-simd': { 'args': None, 'default': False },
#        '-unroll': { 'args': '[|=\d+]', 'default': False, 'space': False },
#        '-vec': { 'args': None, 'default': True },
#        '-no-vec': { 'args': None, 'default': False },
#        '-qopt-report': { 'args': '[|=0|=1|=2|=3|=4|=5]', 'default': False, 'space': False },
#        '-parallel': { 'args': None, 'default': False },
#        '-qopenmp': { 'args': None, 'default': False },
#        '-qno-openmp': { 'args': None, 'default': True },
#        '-fma': { 'args': None, 'default': True },
#        '-no-fma': { 'args': None, 'default': False },
#        '-fp-model': { 'args': '[precise|strict|source|fast=1|fast=2|except|no-except]', 'default': 'fast=1', 'space': True },
#        '-ftz': { 'args': None, 'default': True },
#        '-no-ftz': { 'args': None, 'default': False },
#        '-prec-div': { 'args': None, 'default': True },
#        '-no-prec-div': { 'args': None, 'default': False },
#        '-prec-sqrt': { 'args': None, 'default': True },
#        '-no-prec-sqrt': { 'args': None, 'default': False },
#        '-finline': { 'args': None, 'default': False },
#        '-fno-inline': { 'args': None, 'default': True },
#
#    }
#    pass
#
